"""Implements backpropagation through a while loop by using checkpointing.

(Variously known as "treeverse", "optimal checkpointing", "binomial checkpointing",
"recursive checkpointing", "revolve", etc.)

The algorithm used here is the online version (when the number of steps isn't known in
advance), as proposed in:

    Stumm and Walther 2010
    New Algorithms for Optimal Online Checkpointing
    https://tu-dresden.de/mn/math/wir/ressourcen/dateien/forschung/publikationen/pdf2010/new_algorithms_for_optimal_online_checkpointing.pdf

and also depends on the results of:

    Wang and Moin 2008
    Minimal repetition dynamic checkpointing algorithm for unsteady adjoint calculation
    https://web.stanford.edu/group/ctr/ResBriefs08/4_checkpointing.pdf

This matches the performance of the offline version (classical treeverse, when the
number of steps is known in advance) provided that the number of steps is less than or
equal to `(num_checkpoints + 1) * (num_checkpoints + 2) / 2`; see the Stumm--Walther
paper. After that is may make extra steps (as compared to the offline version), but does
still have similar asymptotic complexity.

For context, the two classical references for (offline) treeverse are:

    Griewank 1992
    Achieiving logarithmic growth of temporal and spatial complexity in reverse
    automatic differentiation
    https://ftp.mcs.anl.gov/pub/tech_reports/reports/P228.pdf

and

    Griewank and Walther 2000
    Algorithm 799: revolve: an implementation of checkpointing for the reverse or
    adjoint mode of computational differentiation
    https://dl.acm.org/doi/pdf/10.1145/347837.347846
"""
# I think this code is not *quite* maximally efficient. A few things that could be
# improved:
# - The initial value is available on the backward pass twice: once as an argument,
#   once as a saved checkpoint. We should be able to get away without this repetition.
# - We only implement Algorithm I of Stumm--Wather. Additionally implementing
#   Algorithm II would be worthwhile. (But finickity, as their description of it in the
#   paper leaves something to be desired. And may also have an off-by-one-error, like
#   their Figure 2.2 does?)

import functools as ft
import math
import operator
import typing
from collections.abc import Callable, Sequence
from typing import Any, cast, TypeVar

import jax
import jax.core
import jax.lax as lax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, ArrayLike, Bool

from ..._ad import filter_closure_convert, filter_custom_jvp, filter_custom_vjp
from ..._errors import error_if
from ..._filters import combine, filter, is_array, is_inexact_array, partition
from ..._module import Static
from ..._tree import tree_at, tree_equal
from .._nontraceable import nonbatchable
from .common import common_rewrite, fixed_asarray


_T = TypeVar("_T")
_Bool = bool | Bool[Array, ""]
_Node = Any


def checkpointed_while_loop(
    cond_fun: Callable[[_T], _Bool],
    body_fun: Callable[[_T], _T],
    init_val: _T,
    *,
    max_steps: int | None = None,
    buffers: Callable[[_T], _Node | Sequence[_Node]] | None = None,
    checkpoints: int | None = None,
) -> _T:
    """Reverse-mode autodifferentiable while loop, using optimal online checkpointing.

    The usual `jax.lax.while_loop` is not reverse-mode autodifferentiable, since it
    would need to save a potentially unbounded amount of residuals between the forward
    and backward pass. However, JAX/XLA requires that all memory buffers be of known
    (bounded) size.

    This works around this limitation by saving values to a prespecified number of
    checkpoints, and then recomputing other intermediate value on-the-fly.

    Checkpointing in this way is a classical autodifferentiation technique, usually used
    to reduce memory consumption. (And it's still useful for this purpose for us too.)

    **Arguments:**

    - `cond_fun`: As `lax.while_loop`.
    - `body_fun`: As `lax.while_loop`.
    - `init_val`: As `lax.while_loop`.
    - `max_steps`: A bound on the maximum number of steps. Set to `None` to allow an
        arbitrary number of steps. (`checkpointed_while_loop` is reverse-mode
        autodifferentiable regardless of whether `max_steps` is finite.)
    - `checkpoints`: The number of steps at which to checkpoint. The memory consumed
        will be that of `checkpoints`-many copies of `init_val`. (As the state is
        updated throughout the loop.)
    - `buffers`: If passed, then every array in `tree_leaves(buffers(init_val))` will
        become a write-only buffer. (Supporting only `.at[].set()`.)

    **Returns:**

    The final value; as `lax.while_loop`.

    !!! Info

        This function is not forward-mode autodifferentiable.

    !!! Info

        `buffers` is useful in the following way.

        Recall that a checkpointed while loop works by storing a copy of the evolving
        value (of the same structure as `init_val`). However, if part of that value is
        a large buffer than is being progressively written in to as the loop progresses,
        then we will end up with multiple copies of this large buffer.

        As this slows things down, we offer a special API specifically to handle this
        case.

    !!! Danger

        Note that `buffers` is subject to the following restrictions:

        - You should never write to the same location twice.
        - You should only read from it (`buf[i]`) at locations (`i`) that you have
          written to previously.

        These assumptions are *completely unchecked* and you will get incorrect
        gradients if you violate these assumptions.

    ??? cite "References"

        Selecting which steps at which to save checkpoints (and when this is done, which
        old checkpoint to evict) is important for minimising the amount of recomputation
        performed.

        This is a difficult, but solved, problem! So if you are using this function in
        academic work, then you should cite the following references.

        The implementation here performs "online checkpointing", as the number of steps
        is not known in advance. This was developed in:

        ```bibtex
        @article{stumm2010new,
            author = {Stumm, Philipp and Walther, Andrea},
            title = {New Algorithms for Optimal Online Checkpointing},
            journal = {SIAM Journal on Scientific Computing},
            volume = {32},
            number = {2},
            pages = {836--854},
            year = {2010},
            doi = {10.1137/080742439},
        }

        @article{wang2009minimal,
            author = {Wang, Qiqi and Moin, Parviz and Iaccarino, Gianluca},
            title = {Minimal Repetition Dynamic Checkpointing Algorithm for Unsteady
                     Adjoint Calculation},
            journal = {SIAM Journal on Scientific Computing},
            volume = {31},
            number = {4},
            pages = {2549--2567},
            year = {2009},
            doi = {10.1137/080727890},
        }
        ```

        For reference, the classical "offline checkpointing" (also known as "treeverse",
        "recursive binary checkpointing", "revolve" etc.) was developed in:

        ```bibtex
        @article{griewank1992achieving,
            author = {Griewank, Andreas},
            title = {Achieving logarithmic growth of temporal and spatial complexity in
                     reverse automatic differentiation},
            journal = {Optimization Methods and Software},
            volume = {1},
            number = {1},
            pages = {35--54},
            year  = {1992},
            publisher = {Taylor & Francis},
            doi = {10.1080/10556789208805505},
        }

        @article{griewank2000revolve,
            author = {Griewank, Andreas and Walther, Andrea},
            title = {Algorithm 799: Revolve: An Implementation of Checkpointing for the
                     Reverse or Adjoint Mode of Computational Differentiation},
            year = {2000},
            publisher = {Association for Computing Machinery},
            volume = {26},
            number = {1},
            doi = {10.1145/347837.347846},
            journal = {ACM Trans. Math. Softw.},
            pages = {19--45},
        }
        ```
    """
    if checkpoints is None:
        if max_steps is None:
            raise ValueError(
                "Must specify either `max_steps` or `checkpoints` in "
                "`equinox.internal.checkpointed_while_loop`."
            )
        # Binomial logarithmic growth is what is needed in classical treeverse.
        #
        # If
        # `max_steps < (checkpoints + 1)(checkpoints + 2)/2`
        # then the time spend recomputing will be optimised; see equation (2.2) of
        # "New Algorithms for Optimal Online Checkpointing", Stumm and Walther 2010.
        # https://tu-dresden.de/mn/math/wir/ressourcen/dateien/forschung/publikationen/pdf2010/new_algorithms_for_optimal_online_checkpointing.pdf
        # and also the `_should_save_residuals` function below.
        #
        # So by default we use `checkpoints = O(sqrt(max_steps))`.
        # (Classical treeverse uses only `O(log(max_steps))`, of course, but doesn't
        # handle the online case.)
        if max_steps == 1:
            checkpoints = 1
        else:
            checkpoints = math.floor(-1.5 + math.sqrt(2 * max_steps + 0.25)) + 1
            assert 2 * max_steps < ((checkpoints + 1) * (checkpoints + 2))
    if checkpoints < 1:
        raise ValueError("Must have at least one checkpoint")
    init_val = jtu.tree_map(fixed_asarray, init_val)
    if max_steps == 0:
        return init_val
    if max_steps is not None:
        checkpoints = min(checkpoints, max_steps)
    cond_fun_, body_fun_, init_val_, buffers_ = common_rewrite(
        cond_fun, body_fun, init_val, max_steps, buffers, makes_false_steps=False
    )
    del cond_fun, body_fun, init_val, buffers
    cond_fun_ = filter_closure_convert(cond_fun_, init_val_)
    cond_fun_ = jtu.tree_map(_stop_gradient, cond_fun_)
    body_fun_ = filter_closure_convert(body_fun_, init_val_)
    vjp_arg = (init_val_, body_fun_)
    final_val_ = _checkpointed_while_loop(
        vjp_arg, cond_fun_, checkpoints, buffers_, max_steps
    )
    _, _, _, final_val = _stop_gradient_on_unperturbed(init_val_, final_val_, body_fun_)
    return final_val


def _stop_gradient(x):
    if is_array(x):
        return lax.stop_gradient(x)
    else:
        return x


@filter_custom_vjp
def _checkpointed_while_loop(vjp_arg, cond_fun, checkpoints, buffers, max_steps):
    """Uncheckpointed forward used when not differentiating."""
    del checkpoints, buffers, max_steps
    init_val, body_fun = vjp_arg
    while_loop = jax.named_call(lax.while_loop, name="checkpointed-no-vjp")
    # Hashable wrapper; JAX issue #13554 and
    # https://github.com/patrick-kidger/equinox/issues/768
    return while_loop(lambda x: cond_fun(x), lambda x: body_fun(x), init_val)


def _scalar_index(i, x):
    """As `x[i]`, but slightly more efficient for a nonnegative scalar `i`.

    (As it avoids support for negative indexing, and lowers to `dynamic_slice` rather
     than `gather`.)
    """
    assert jnp.shape(i) == ()
    return lax.dynamic_index_in_dim(x, i, keepdims=False)


def _unique_index(i, x):
    """As `x[i]`, but states that `i` has unique indices."""
    return x.at[i].get(unique_indices=True)


def _stumm_walther_i(step, save_state):
    """Algorithm 1 from:

    Stumm and Walther 2010
    New Algorithms for Optimal Online Checkpointing
    https://tu-dresden.de/mn/math/wir/ressourcen/dateien/forschung/publikationen/pdf2010/new_algorithms_for_optimal_online_checkpointing.pdf
    """
    step, save_state = nonbatchable((step, save_state))
    i, o, p, s = save_state
    i = cast(ArrayLike, i)
    s = cast(ArrayLike, s)
    index = i
    save_residual = s
    i = jnp.where(s, i + 1, i)
    i = jnp.where(s & (i > o), 1, i)
    s = jnp.where(step + 1 == p, False, s)
    pred = step == p
    p = jnp.where(pred, p + o, p)
    o = jnp.where(pred, o - 1, o)
    i = jnp.where(pred, o, i)
    s = jnp.where(pred, o > 0, s)
    out = save_residual, index, (i, o, p, s)
    msg = (
        "Internal run-time error when checkpointing "
        "`equinox.internal.checkpointed_while_loop`. "
        "Please raise an issue at https://github.com/patrick-kidger/equinox"
    )
    if getattr(typing, "TESTING", False):
        out = error_if(out, pred & (o == -1), msg)
    out = nonbatchable(out)
    return out


def _any_dispensable(dispensable, residual_steps: Array, levels):
    del levels
    dispensable_steps = jnp.where(dispensable, residual_steps, 0)
    index = dispensable_steps.argmax()
    level = 0
    dispensable2 = dispensable.at[index].set(False)
    return index, level, dispensable2


def _none_dispensable(dispensable, residual_steps, levels):
    index = residual_steps.argmax()
    level = levels[index] + 1
    dispensable2 = jnp.where(levels < levels[index], True, dispensable)
    return index, level, dispensable2


def _wang_moin(step, save_state, residual_steps):
    """Algorithm 1 or 3 from:

    Wang and Moin 2008
    Minimal repetition dynamic checkpointing algorithm for unsteady adjoint calculation
    https://web.stanford.edu/group/ctr/ResBriefs08/4_checkpointing.pdf
    """
    step, save_state, residual_steps = nonbatchable((step, save_state, residual_steps))
    levels, dispensable = save_state
    if len(residual_steps) == 1:
        # Don't save if we only have space to save the initial value, which is already
        # stored.
        save_residual = False
        index = 0
        levels2 = levels
        dispensable2 = dispensable
    else:
        save_residual = len(residual_steps) > 1
        index, level, dispensable2 = lax.cond(
            dispensable.any(),
            _any_dispensable,
            _none_dispensable,
            dispensable,
            residual_steps,
            levels,
        )
        levels2 = levels.at[index].set(level)
    out = save_residual, index, (levels2, dispensable2)
    out = nonbatchable(out)
    return out


def _stumm_walther_i_wrapper(step, save_state, residual_steps):
    del residual_steps
    save_state_sw_i, save_state_wm = save_state
    save_residual, index, save_state_sw_i_2 = _stumm_walther_i(step, save_state_sw_i)
    return save_residual, index, (save_state_sw_i_2, save_state_wm)


def _wang_moin_wrapper(step, save_state, residual_steps):
    save_state_sw_i, save_state_wm = save_state
    save_residual, index, save_state_wm_2 = _wang_moin(
        step, save_state_wm, residual_steps
    )
    return save_residual, index, (save_state_sw_i, save_state_wm_2)


def _should_save_residual(
    step, save_state, residual_steps, u2_minus_1, checkpoints, max_steps
):
    """This is the controller for whether we should save the current value at each step,
    and if so which memory location to save it in.
    """
    # TODO: also implement Algorithm 2 of Stumm and Walther, which gives improved
    # results for u2 < step < u3.

    if checkpoints == max_steps:
        # Important special case! We don't need to do any computations in this case.
        # This is a measurable performance optimisation.
        save_residual = True
        index = step
        save_state2 = save_state
    else:
        if max_steps is not None and max_steps <= u2_minus_1:
            # `step < max_steps` implies `step + 1 <= max_steps`. If it is also the case
            # that `max_steps <= u2 - 1` then overall we know that `step <= u2 - 2`,
            # which is the condition for using Stumm--Walther.
            #
            # Skipping this cond offers a measurable performance optimisation.
            pred = False
        else:
            step, u2_minus_1 = nonbatchable((step, u2_minus_1))
            pred = step >= u2_minus_1
        save_residual, index, save_state2 = lax.cond(
            pred,
            _wang_moin_wrapper,
            _stumm_walther_i_wrapper,
            step,
            save_state,
            residual_steps,
        )
    return save_residual, index, save_state2


def _unreachable_checkpoint_step(x):
    """Dummy value used to represent a checkpoint we never reach."""
    dtype = jnp.result_type(x)  # x can be a dtype or an arraylike
    return jnp.iinfo(dtype).max


def _array_to_none(x):
    assert is_array(x)
    return None


@_checkpointed_while_loop.def_fwd
def _checkpointed_while_loop_fwd(
    perturbed, vjp_arg, cond_fun, checkpoints, buffers, max_steps
):
    """Run the while loop, saving checkpoints whenever the controller
    (`_should_save_residual`) requires.
    """
    del perturbed
    init_val, body_fun = vjp_arg
    # Equation (2.2) of Stumm and Walther
    u2_minus_1 = ((checkpoints + 1) * (checkpoints + 2)) // 2 - 1

    def _cond_fun(carry):
        _, _, val, _, _ = carry
        return cond_fun(val)

    def _body_fun(carry):
        step, save_state, val, residual_steps, residuals = carry
        save_state, residual_steps = nonbatchable((save_state, residual_steps))

        step2 = step + 1
        save_residual, index, save_state2 = _should_save_residual(
            step, save_state, residual_steps, u2_minus_1, checkpoints, max_steps
        )
        val2 = body_fun(val)

        def _maybe_update(xs, x):
            where_x = jnp.where(save_residual, x, _scalar_index(index, xs))
            where_x = cast(Array, where_x)
            return lax.dynamic_update_index_in_dim(xs, where_x, index, axis=0)

        val_no_buffers = tree_at(buffers(None), val, replace_fn=_array_to_none)
        residual_steps2, residuals2 = jtu.tree_map(
            _maybe_update,
            (residual_steps, residuals),
            (step, val_no_buffers),
        )
        save_state2, residual_steps2 = nonbatchable((save_state2, residual_steps2))
        return step2, save_state2, val2, residual_steps2, residuals2

    int_dtype = jnp.int64 if jax.config.jax_enable_x64 else jnp.int32  # pyright: ignore
    init_step = jnp.array(0, dtype=int_dtype)  # dtype matches init_residual_steps
    init_save_state_sw_i = 0, checkpoints, checkpoints, True
    dtype_max = jnp.iinfo(int_dtype).max  # pyright: ignore
    init_save_state_wm = (
        jnp.zeros(checkpoints, dtype=int_dtype).at[0].set(dtype_max),
        jnp.full((checkpoints,), False),
    )
    init_save_state = (init_save_state_sw_i, init_save_state_wm)
    # Uses the fact that `_unreachable_checkpoint_step` returns intmax, so that in our
    # sorting later, in the steps < checkpoints case, all unused memory gets sorted
    # to the end.
    init_residual_steps = jnp.full(
        checkpoints, _unreachable_checkpoint_step(int_dtype), dtype=int_dtype
    )

    init_val_no_buffers = tree_at(buffers(None), init_val, replace_fn=_array_to_none)
    # Fill value for the memory isn't important.
    init_residuals = jtu.tree_map(
        lambda x: jnp.zeros((checkpoints,) + x.shape, x.dtype), init_val_no_buffers
    )
    init_carry = (
        init_step,
        init_save_state,
        init_val,
        init_residual_steps,
        init_residuals,
    )
    # `step` is an increasing counter 0, 1, 2, 3, ...
    # `save_state` is state used by the logic for whether to save a checkpoint each step
    # `val` is the evolving state of the loop
    # `residual_steps` is the buffer of the `step` for each checkpoint
    # `residuals` is the buffer of the `val` for each checkpoint

    while_loop = jax.named_call(lax.while_loop, name="checkpointed-fwd")
    final_carry = while_loop(_cond_fun, _body_fun, init_carry)
    num_steps, _, final_val, final_residual_steps, final_residuals = final_carry
    filled_buffers = buffers(_is_none)(final_val)

    # The above procedure may produce residuals saved in jumbled-up order. Meanwhile
    # treeverse (used on the backward pass) treats the residuals like a stack,
    # reading and writing the most recent residual to and from the end. So sort the
    # residuals we've produced here to obtain the desired invariant, i.e. that the
    # residuals are in order.
    if checkpoints != max_steps:
        # If `checkpoints == max_steps` then residuals are already sorted.
        sort_indices = jnp.argsort(final_residual_steps)
        final_residual_steps, final_residuals = jtu.tree_map(
            ft.partial(_unique_index, sort_indices),
            (final_residual_steps, final_residuals),
        )
    num_steps, final_residual_steps = nonbatchable((num_steps, final_residual_steps))
    return final_val, (num_steps, final_residual_steps, final_residuals, filled_buffers)


def _load_from_checkpoint(
    step_grad_val, index, residual_steps, residuals, checkpoints, max_steps
):
    """Loads a residual from the store of checkpoints."""
    # step_grad_val is the current location of grad_val.
    # index is the next currently empty slot for saving a residual.
    step_grad_val, index, residual_steps = nonbatchable(
        (step_grad_val, index, residual_steps)
    )

    # Subtract one to get the most recent residual, and then load it.
    # (Clip to zero just to not error from index == 0 on the very last step; the result
    # is unused in this case.)
    read_index = jnp.maximum(index - 1, 0)
    step_val2, val2 = jtu.tree_map(
        ft.partial(_scalar_index, read_index), (residual_steps, residuals)
    )

    # We may need to keep this residual around, and jump back to it multiple times.
    # (In which case index2 == index.) Or this may be the last time and we don't need to
    # save it any more. (In which case index2 == index - 1.)
    # If `step_val2 + 1 == step_grad_val2` then we're about to make a U-turn on the next
    # step, so we won't need to load from this checkpoint. (And as
    # `_load_from_checkpoint` is itself used within a U-turn, then in practice this
    # triggers whenever we get >1 U-turns back-to-back.)
    if checkpoints == max_steps:
        # Known to be symbolically true in this case; this is a slight performance
        # optimisation.
        pred = True
    else:
        pred = step_val2 + 1 == step_grad_val
    index2 = jnp.where(pred, read_index, index)
    step_val2, index2 = nonbatchable((step_val2, index2))
    return step_val2, val2, index2


def _maybe_save_to_checkpoint(
    step_val,
    step_grad_val,
    step_next_checkpoint,
    index,
    val,
    residual_steps,
    residuals,
    checkpoints,
):
    """Might save a residual to the store of checkpoints."""
    (
        step_val,
        step_grad_val,
        step_next_checkpoint,
        index,
        residual_steps,
    ) = nonbatchable(
        (step_val, step_grad_val, step_next_checkpoint, index, residual_steps)
    )
    save_checkpoint = step_val == step_next_checkpoint

    def _maybe_update(xs, x):
        where_x = jnp.where(save_checkpoint, x, _scalar_index(index, xs))
        where_x = cast(Array, where_x)
        return lax.dynamic_update_index_in_dim(xs, where_x, index, axis=0)

    residual_steps2, residuals2 = jtu.tree_map(
        _maybe_update, (residual_steps, residuals), (step_val, val)
    )
    index2 = jnp.where(save_checkpoint, index + 1, index)
    step_next_checkpoint2 = jnp.where(
        save_checkpoint,
        _calc_next_checkpoint(step_val, step_grad_val, index2, checkpoints),
        step_next_checkpoint,
    )
    index2, step_next_checkpoint2, residual_steps2 = nonbatchable(
        (index2, step_next_checkpoint2, residual_steps2)
    )
    return index2, step_next_checkpoint2, residual_steps2, residuals2


def _calc_next_checkpoint(step_val, step_grad_val, index, checkpoints):
    """Determines the step at which we next want to save a checkpoint."""
    # Note that when this function is called, `step_val` is always at the most recent
    # checkpoint.
    step_val, step_grad_val, index = nonbatchable((step_val, step_grad_val, index))

    # Using treeverse...
    # ...Checkpoints are either placed binomially (most of the time)...
    out_binomial = step_val + (step_grad_val - step_val) // 2
    # ...or linearly (when the space to cross fits within the checkpoint budget).
    out_linear = step_val + 1
    within_budget = (step_grad_val - step_val - 2) <= (checkpoints - index)
    out = jnp.where(within_budget, out_linear, out_binomial)
    # Why -2?
    # If `step_val + 1 == step_grad_val` then we're just going to make a single U-turn,
    # and don't need to store any checkpoints.
    # If `step_val + 2 == step_grad_val` then we're just going make a fwd, then a
    # U-turn, and can then load the existing checkpoint at `step_val` in order to make
    # the next U-turn. Once again we don't need to store any checkpoints.
    # If `step_val + 3 == step_grad_val` then ideally we would store a checkpoint at
    # `step_val + 1`.
    # Meanwhile, `checkpoints - index` is the number of spaces we have left in which to
    # store checkpoints. (e.g. `checkpoints == index` indicates that our buffer is
    # full.)
    # Thus if `step_val + 3 == step_grad_val` then
    # `step_grad_val - step_val - 2 <= checkpoints - index`
    # is the desired condition. (For "save a checkpoint if you can".)
    # Now proceed by induction: if `step_val + i == step_grad_val` for some `i > 3` and
    # we trigger this condition, we will reduce the LHS by one (as `step_val `increments
    # in a fwd step) and increase the RHS side by one (as `index` increments as we save
    # a checkpoint) and we eventually reduce back to the `step_val + 3 == step_grad_val`
    # case.

    # Logic as above: in these cases we don't need to store a checkpoint at all.
    no_checkpoint = index == checkpoints
    no_checkpoint = no_checkpoint | (step_val + 1 == step_grad_val)
    no_checkpoint = no_checkpoint | (step_val + 2 == step_grad_val)

    step_next_checkpoint = jnp.where(
        no_checkpoint, _unreachable_checkpoint_step(out), out
    )
    # Invariant: `step_val < step_next_checkpoint`. (Due to the invariant
    # `step_val + 1 <= step_grad_val`.)
    step_next_checkpoint = nonbatchable(step_next_checkpoint)
    return step_next_checkpoint


def _is_none(x):
    return x is None


def _get_value(x):
    assert type(x) is jax.custom_derivatives.CustomVJPPrimal
    return x.value


def _is_symbolic_zero_or_none(x):
    return isinstance(x, jax.custom_derivatives.SymbolicZero) or x is None


def _not_symbolic_zero(x):
    return not isinstance(x, jax.custom_derivatives.SymbolicZero)


def _materialise_symbolic_zero(is_zero, ct_x, x):
    if is_zero:
        assert ct_x is None
        return None
    else:
        if ct_x is None:
            assert is_inexact_array(x)
            return jnp.zeros_like(x)
        else:
            return ct_x


def _assert_bool(x):
    assert type(x) is bool


def _fwd(
    step_val,
    step_grad_val,
    step_next_checkpoint,
    index,
    val_candidate,
    grad_val,
    grad_body_fun,
):
    """Propagates the primal forward one step."""
    step_val2 = nonbatchable(step_val + 1)
    return (
        step_val2,
        step_grad_val,
        step_next_checkpoint,
        index,
        val_candidate,
        grad_val,
        grad_body_fun,
    )


def _make_u_turn(vjp_fn, residual_steps, residuals, checkpoints, max_steps):
    """Propagates the cotangent backward one step."""
    residual_steps = nonbatchable(residual_steps)

    def _u_turn(
        step_val,
        step_grad_val,
        step_next_checkpoint,
        index,
        val_candidate,
        grad_val,
        grad_body_fun,
    ):
        del step_val, step_next_checkpoint
        step_grad_val, index = nonbatchable((step_grad_val, index))
        grad_body_fun_update, grad_val2 = vjp_fn(grad_val)
        grad_body_fun2 = jtu.tree_map(operator.add, grad_body_fun, grad_body_fun_update)
        step_grad_val2 = step_grad_val - 1
        step_val2, val2, index2 = _load_from_checkpoint(
            step_grad_val2, index, residual_steps, residuals, checkpoints, max_steps
        )
        step_next_checkpoint2 = _calc_next_checkpoint(
            step_val2, step_grad_val2, index2, checkpoints
        )
        step_val2, step_grad_val2, step_next_checkpoint2, index2 = nonbatchable(
            (step_val2, step_grad_val2, step_next_checkpoint2, index2)
        )
        return (
            step_val2,
            step_grad_val2,
            step_next_checkpoint2,
            index2,
            val2,
            grad_val2,
            grad_body_fun2,
        )

    return _u_turn


@_checkpointed_while_loop.def_bwd
def _checkpointed_while_loop_bwd(
    remainders,
    grad_final_val,
    perturbed,
    vjp_arg,
    cond_fun,
    checkpoints,
    buffers,
    max_steps,
):
    """Time for the complicated bit: iterate backward through a checkpointed while loop,
    loading values from checkpoints and using treeverse to toggle between forward and
    backward steps.
    """
    perturb_init_val, perturb_body_fun = perturbed
    _, body_fun = vjp_arg
    grad_final_body_fun = jtu.tree_map(
        lambda x, p: jnp.zeros_like(x) if p else None, body_fun, perturb_body_fun
    )
    del cond_fun, perturbed
    num_steps, init_residual_steps, init_residuals, filled_buffers = remainders
    # `allow_constant_across_batch` to handle
    # https://github.com/patrick-kidger/optimistix/issues/67
    num_steps, init_residual_steps = nonbatchable(
        (num_steps, init_residual_steps), allow_constant_across_batch=True
    )

    def _cond_fun(carry):
        _, step_grad_val, *_ = carry
        # step_grad_val is the location of our cotangent. We want to keep going until
        # this has got all the way to the start.
        step_grad_val = nonbatchable(step_grad_val)
        return step_grad_val > 0

    def _body_fun(carry):
        (
            step_val,
            step_grad_val,
            step_next_checkpoint,
            index,
            val,
            grad_val,
            grad_body_fun,
            residual_steps,
            residuals,
        ) = carry
        (
            step_val,
            step_grad_val,
            step_next_checkpoint,
            index,
            residual_steps,
        ) = nonbatchable(
            (step_val, step_grad_val, step_next_checkpoint, index, residual_steps)
        )

        msg = (
            "Internal run-time error when backpropagating through "
            "`equinox.internal.checkpointed_while_loop`. "
            "Please raise an issue at https://github.com/patrick-kidger/equinox"
        )
        if getattr(typing, "TESTING", False):
            step_val = error_if(step_val, step_val >= step_grad_val, msg)

        #
        # First either propagate our primal state forward, or make a U-turn if the
        # primal state has caught up to the cotangent state.
        #

        # Use `filter_vjp` to neatly handle floating-point arrays.
        #
        # We pass in `body_fun` as an argument as it contains its closed-over values
        # in its PyTree structure, and we do want to compute cotangents wrt these.
        val_buffers = tree_at(buffers(_is_none), val, filled_buffers, is_leaf=_is_none)
        grad_val_has_ct = jtu.tree_map(is_array, grad_val, is_leaf=_is_none)
        body_fun_has_ct = jtu.tree_map(is_array, grad_body_fun, is_leaf=_is_none)
        diff_val, nondiff_val = partition(val_buffers, grad_val_has_ct)
        diff_body, nondiff_body = partition(body_fun, body_fun_has_ct)

        def _to_vjp(_diff_body, _diff_val):
            _val_buffers = combine(_diff_val, nondiff_val)
            _body_fun = combine(_diff_body, nondiff_body)
            _out = _body_fun(_val_buffers)
            return partition(_out, grad_val_has_ct)

        dynamic_val_candidate, vjp_fn, static_val_candidate = jax.vjp(
            _to_vjp, diff_body, diff_val, has_aux=True
        )
        val_candidate = combine(dynamic_val_candidate, static_val_candidate)
        val_candidate = tree_at(buffers(None), val_candidate, replace_fn=lambda _: None)

        if checkpoints == max_steps:
            # Skip calculating this at runtime; we know it must be true.
            # This means we can inline the U-turn branch of the `cond`.
            perform_u_turn = True
        else:
            perform_u_turn = step_val + 1 == step_grad_val
            perform_u_turn = nonbatchable(perform_u_turn)
        (
            step_val2,
            step_grad_val2,
            step_next_checkpoint2,
            index2,
            val2,
            grad_val2,
            grad_body_fun2,
        ) = lax.cond(
            perform_u_turn,
            _make_u_turn(vjp_fn, residual_steps, residuals, checkpoints, max_steps),
            _fwd,
            step_val,
            step_grad_val,
            step_next_checkpoint,
            index,
            val_candidate,
            grad_val,
            grad_body_fun,
        )

        #
        # Second, decide whether to store our current primal state in a checkpoint.
        # Note that this can only actually trigger on `_fwd` and not on `_u_turn`, as
        # if `_u_turn` happens then `step_val2 < step_next_checkpoint2`, but
        # `_maybe_save_to_checkpoint` has a `step_val2 == step_next_checkpoint2` check.
        # (We could maybe move this inside `_fwd`? It was originally outside for
        # efficiency reasons that I don't think apply any more.)
        #

        if checkpoints == max_steps:
            # In this case we never need to save any extra checkpoints on the backward
            # pass, so skip this.
            residual_steps2 = residual_steps
            residuals2 = residuals
        else:
            (
                index2,
                step_next_checkpoint2,
                residual_steps2,
                residuals2,
            ) = _maybe_save_to_checkpoint(
                step_val2,
                step_grad_val2,
                step_next_checkpoint2,
                index2,
                val2,
                residual_steps,
                residuals,
                checkpoints,
            )

        return (
            step_val2,
            step_grad_val2,
            step_next_checkpoint2,
            index2,
            val2,
            grad_val2,
            grad_body_fun2,
            residual_steps2,
            residuals2,
        )

    # We can index into our residuals using 0, 1, ..., checkpoints - 1.
    # `index` is used to refer to the next empty spot, so it takes values in
    # 0, 1, ..., checkpoints - 1, checkpoints, where `index == checkpoints` indicates
    # that there are no empty spots and the whole buffer is full. (And this is used in
    # `_calc_step_next_checkpoint`.)
    init_index = jnp.minimum(num_steps, checkpoints)
    init_step_grad_val = num_steps
    init_step_val, init_val, init_index = _load_from_checkpoint(
        init_step_grad_val,
        init_index,
        init_residual_steps,
        init_residuals,
        checkpoints,
        max_steps,
    )
    init_step_next_checkpoint = _calc_next_checkpoint(
        init_step_val, init_step_grad_val, init_index, checkpoints
    )

    #
    # Okay, this next bit is a bit complicated.
    # We need to figure out exactly which parts of the carry that we're going to compute
    # cotangents for.
    # There are three interacting pieces:
    # (1) Only some inputs even need their cotangents computed (those that are marked
    #     as perturbed). We should avoid computing intermediate cotangents that aren't
    #     needed.
    # (2) Some cotangents may be symbolic zeros. We should avoid materialising any that
    #     don't need to be materialised.
    # (3) Some parts of the body function may not be differentiable. (E.g. they call to
    #     `eqxi.nondifferentiable` or `eqxi.nondifferentiable_backward`.) In particular,
    #     some inexact arrays may be marked with these. (Anything else is automatically
    #     nondifferentiable.) As these raise an error eagerly, we can't just take an
    #     approach of "differentiate everything, then DCE".
    #
    # Our strategy is as follows. First, consider those inputs that are perturbed, and
    # iterate to find all the carries that become perturbed as a result. We obtain a
    # fixed point.
    # * This mimics what JAX does in JVP rules for scans/whiles. And in fact we also
    #   track perturbation information forward via JVPs (which input tangents affect
    #   which output tangents), rather than backward via VJPs (which output cotangents
    #   affect which input cotangents), as this is what JAX does here and we'd like to
    #   be consistent. (Also, it seems like there might be a catch-22 trying to do this
    #   with VJPs -- how do you decide which inputs to perturb in your VJP?) This is
    #   basically handling criteria (1).
    # * This also means that we're perturbing the minimal amount possible. We don't even
    #   run the JVP rule for something that isn't differentiated. This is part of
    #   addressing criteria (3).
    #
    # Second, we ignore all cotangents (of our forward output) that aren't on a
    # perturbed carry. Clearly they're not going to affect anything. So we switch them
    # to symbolic zeros.
    # * This is part of addressing criteria (3) -- we avoid propagating cotangents
    #   that we don't need to. Maybe there's an `eqxi.nondifferentiable_backward` in
    #   there.
    #
    # Third, we find another fixed point, this time for the symbolic zeros. Track which
    # cotangents affect which cotangents, and materialise the minimal amount possible.
    # * This addresses criteria (2).
    # * Note that symbolic zeros are doing double-duty here. We avoid materialising them
    #   for simple efficiency reasons, sure -- but we also want to keep as many
    #   unmaterialised as possible, in case there's an `eqxi.nondifferentiable_backward`
    #   in there, which only accepts symbolic zero cotangents.
    #
    # Phew! At that point, job done: we know which cotangents we're propagating and so
    # we can run our loop.
    #
    # The first fixed point basically mimics what JAX does in the JVP rule for a while
    # or scan, figuring out which values are perturbed. The second fixed point basically
    # mimics what JAX does in the transpose rule of a scan, figuring out which
    # cotangents are symbolic zeros.
    #

    # Do this inside the dynamic context of a `jax.eval_shape` so that it's as cheap as
    # possible, without extra tracing.
    def _resolve_perturb_val():
        # TODO: is this function still necessary now that we have
        # _stop_gradient_on_unperturbed_jvp?
        init_val_buffers = tree_at(
            buffers(_is_none), init_val, filled_buffers, is_leaf=_is_none
        )
        perturb_val = perturb_init_val
        assert jtu.tree_structure(init_val_buffers) == jtu.tree_structure(perturb_val)

        while True:
            # `body_fun` is included so that we also track perturbatations on that.
            perturb_pair = (perturb_body_fun, perturb_val)
            dynamic, static = partition((body_fun, init_val_buffers), perturb_pair)
            new_perturb_val = sentinel = object()

            @jax.custom_jvp
            def _record_symbolic_zeros(_dynamic_out):
                return _dynamic_out

            def _record_symbolic_zeros_jvp(primals, tangents):
                (primals,) = primals
                (tangents,) = tangents
                nonlocal new_perturb_val
                new_perturb_val = jtu.tree_map(_not_symbolic_zero, tangents)
                return primals, tangents

            _record_symbolic_zeros.defjvp(
                _record_symbolic_zeros_jvp, symbolic_zeros=True
            )

            def _to_linearize(_dynamic):
                _body_fun, _val = combine(_dynamic, static)
                _out = _body_fun(_val)
                _dynamic_out, _static_out = partition(_out, is_inexact_array)
                _dynamic_out = _record_symbolic_zeros(_dynamic_out)
                _out = combine(_dynamic_out, _static_out)
                return _out

            # Not `jax.jvp`, so as not to error if `body_fun` has any `custom_vjp`s.
            jax.linearize(_to_linearize, dynamic)
            assert new_perturb_val is not sentinel
            assert jtu.tree_structure(
                perturb_val, is_leaf=_is_none
            ) == jtu.tree_structure(new_perturb_val, is_leaf=_is_none)
            new_perturb_val = jtu.tree_map(
                lambda x, y: False if (x is False and y is None) else y,
                perturb_val,
                new_perturb_val,
            )
            assert jtu.tree_structure(perturb_val) == jtu.tree_structure(
                new_perturb_val
            )
            if tree_equal(perturb_val, new_perturb_val):
                jtu.tree_map(_assert_bool, perturb_val)
                if getattr(typing, "TESTING", False):
                    print("perturb_val", perturb_val)
                return Static(perturb_val)
            else:
                perturb_val = jtu.tree_map(operator.or_, perturb_val, new_perturb_val)

    # Find a fixed point of which values we need to perturb. (Not the same as whether
    # or not they have symbolic zero cotangents! All unperturbed values must have
    # symbolic zero cotangents, but some perturbed values may have these as well.)
    perturb_val = jax.eval_shape(_resolve_perturb_val).value

    # Do this inside the dynamic context of a `jax.eval_shape` so that it's as cheap as
    # possible, without extra tracing.
    def _resolve_symbolic_zeros(grad_val):
        symbolic_zero_gradient = jtu.tree_map(_is_none, grad_val, is_leaf=_is_none)
        init_val_buffers = tree_at(
            buffers(_is_none), init_val, filled_buffers, is_leaf=_is_none
        )
        dynamic_init_val, static_init_val = partition(init_val_buffers, perturb_val)

        while True:
            new_symbolic_zero_gradient = sentinel = object()

            # Some hackery to extract which inputs still have symbolic zero cotangents.
            @jax.custom_vjp
            def _record_symbolic_zero(_dynamic_val):
                return _dynamic_val

            def _record_symbolic_zero_fwd(_dynamic_val):
                return jtu.tree_map(_get_value, _dynamic_val), None

            def _record_symbolic_zero_bwd(_, grad_dynamic_val):
                nonlocal new_symbolic_zero_gradient
                new_symbolic_zero_gradient = jtu.tree_map(
                    _is_symbolic_zero_or_none, grad_dynamic_val, is_leaf=_is_none
                )
                return (grad_dynamic_val,)

            _record_symbolic_zero.defvjp(
                _record_symbolic_zero_fwd,
                _record_symbolic_zero_bwd,
                symbolic_zeros=True,
            )

            def _to_vjp(_dynamic_val):
                _dynamic_val = _record_symbolic_zero(_dynamic_val)
                val = combine(_dynamic_val, static_init_val)
                return filter(body_fun(val), symbolic_zero_gradient, inverse=True)

            _, vjp_fn = jax.vjp(_to_vjp, dynamic_init_val)
            vjp_fn(grad_val)
            assert new_symbolic_zero_gradient is not sentinel
            if tree_equal(symbolic_zero_gradient, new_symbolic_zero_gradient):
                jtu.tree_map(_assert_bool, symbolic_zero_gradient)
                if getattr(typing, "TESTING", False):
                    print("symbolic_zero_gradient", symbolic_zero_gradient)
                return Static(symbolic_zero_gradient)
            else:
                symbolic_zero_gradient = jtu.tree_map(
                    operator.and_, symbolic_zero_gradient, new_symbolic_zero_gradient
                )
                grad_val = jtu.tree_map(
                    _materialise_symbolic_zero,
                    symbolic_zero_gradient,
                    grad_val,
                    init_val_buffers,
                )

    grad_final_val = filter(grad_final_val, perturb_val)
    # If all provided cotangents end up being irrelevant -- i.e. the perturbed inputs
    # depend only on those outputs which have symbolic zero cotangents -- then we can
    # skip the whole computation.
    if len(jtu.tree_leaves(grad_final_val)) == 0:
        return jtu.tree_map(lambda _: None, (grad_final_val, body_fun))
    symbolic_zero_gradient = jax.eval_shape(
        _resolve_symbolic_zeros, grad_final_val
    ).value
    init_val_buffers = tree_at(
        buffers(_is_none), init_val, filled_buffers, is_leaf=_is_none
    )
    grad_final_val = jtu.tree_map(
        _materialise_symbolic_zero,
        symbolic_zero_gradient,
        grad_final_val,
        init_val_buffers,
    )

    init_carry = (
        init_step_val,
        init_step_grad_val,
        init_step_next_checkpoint,
        init_index,
        init_val,
        grad_final_val,
        grad_final_body_fun,
        init_residual_steps,
        init_residuals,
    )
    # Note that the saved checkpoints hold both (a) values computed on the forward
    # pass, and (b) checkpoints recomputed on the backward pass. (We don't need to
    # distinguish them.)
    #
    # Controller State
    # ----------------
    # `step_val`: see `val`.
    # `step_grad_val`: see `grad_val`. Note that there is an invariant
    #   `step_val + 1 <= step_grad_val`.
    # `step_next_checkpoint`: the step at which we next need to save a checkpoint, for
    #   the recomputed forward computation.
    # `index` is the index of the memory buffer to save the next checkpoint, for the
    #   recomputed forward computation.
    #
    # Numerical computations
    # ----------------------
    # `val` is the evolving state of the forward loop (reloaded from checkpoints). The
    #   forward loop step that this comes from is `step_val`. As such `val` and
    #   `step_val` will jump back-and-forth as values are loaded from checkpoints and
    #   then recomputed forward using `_fwd`.
    # `grad_val` is the cotangent that we're propagating backward. The step of the
    #   forward loop (that it holds the cotangent for) is given in `step_grad_val`. As
    #   such `step_grad_val` decrements weakly monotonically, and `grad_val` is updated
    #   on every U-turn.
    # `grad_body_fun` is the cotangent being accumulated for `body_fun`. It updates on
    #   every U-turn.
    #
    # Checkpoints
    # -------
    # `residual_steps` is the memory holding the `step` for each checkpoint
    # `residuals` is the memory holding the `val` for each checkpoint

    # Not reverse mode autodifferentiable, but it is forward-mode autodifferentiable!
    # That means we can do Hessians, at least.
    while_loop = jax.named_call(lax.while_loop, name="checkpointed-bwd")
    final_carry = while_loop(_cond_fun, _body_fun, init_carry)
    *_, grad_init_val, grad_body_fun, _, _ = final_carry
    out = grad_init_val, grad_body_fun
    return out


def _resolve_perturb_val(final_val, body_fun, perturb_final_val, perturb_body_fun):
    def _resolve_perturb_val_impl():
        perturb_val = perturb_final_val
        assert jtu.tree_structure(final_val) == jtu.tree_structure(perturb_val)

        while True:
            # `body_fun` is included so that we also track perturbatations on that.
            perturb_pair = (perturb_body_fun, perturb_val)
            dynamic, static = partition((body_fun, final_val), perturb_pair)
            new_perturb_val = sentinel = object()

            @jax.custom_jvp
            def _record_symbolic_zeros(_dynamic_out):
                return _dynamic_out

            def _record_symbolic_zeros_jvp(primals, tangents):
                (primals,) = primals
                (tangents,) = tangents
                nonlocal new_perturb_val
                new_perturb_val = jtu.tree_map(_not_symbolic_zero, tangents)
                return primals, tangents

            _record_symbolic_zeros.defjvp(
                _record_symbolic_zeros_jvp, symbolic_zeros=True
            )

            def _to_linearize(_dynamic):
                _body_fun, _val = combine(_dynamic, static)
                _out = _body_fun(_val)
                _dynamic_out, _static_out = partition(_out, is_inexact_array)
                _dynamic_out = _record_symbolic_zeros(_dynamic_out)
                _out = combine(_dynamic_out, _static_out)
                return _out

            # Not `jax.jvp`, so as not to error if `body_fun` has any `custom_vjp`s.
            jax.linearize(_to_linearize, dynamic)
            if new_perturb_val is sentinel:
                # `_dynamic_out` in `_to_linearize` had no JVP tracers at all, despite
                # `_dynamic` having them. Presumably the user's `_body_fun` has no
                # differentiable dependency whatsoever.
                # This can happen if all the autograd is happening through
                # `perturb_body_fun`.
                return Static(perturb_val)
            assert jtu.tree_structure(
                perturb_val, is_leaf=_is_none
            ) == jtu.tree_structure(new_perturb_val, is_leaf=_is_none)
            new_perturb_val = jtu.tree_map(
                lambda x, y: False if (x is False and y is None) else y,
                perturb_val,
                new_perturb_val,
            )
            assert jtu.tree_structure(perturb_val) == jtu.tree_structure(
                new_perturb_val
            )
            if tree_equal(perturb_val, new_perturb_val):
                jtu.tree_map(_assert_bool, perturb_val)
                if getattr(typing, "TESTING", False):
                    print("stop_gradient_perturb_val", perturb_val)
                return Static(perturb_val)
            else:
                perturb_val = jtu.tree_map(operator.or_, perturb_val, new_perturb_val)

    perturb_val = jax.eval_shape(_resolve_perturb_val_impl).value
    return perturb_val


@filter_custom_jvp
def _stop_gradient_on_unperturbed(init_val, final_val, body_fun):
    del body_fun
    return final_val


def _perturb_to_tang(t, p):
    if t is None:
        return None
    elif p is None:
        return None
    elif p is False:
        return None
    elif p is True:
        return t
    else:
        assert False


@_stop_gradient_on_unperturbed.def_jvp
def _stop_gradient_on_unperturbed_jvp(primals, tangents):
    init_val, final_val, body_fun = primals
    t_init_val, t_final_val, t_body_fun = tangents
    del primals, tangents
    perturb_val, perturb_body_fun = jtu.tree_map(
        lambda _, t: t is not None, (init_val, body_fun), (t_init_val, t_body_fun)
    )
    perturb_val = _resolve_perturb_val(
        init_val, body_fun, perturb_val, perturb_body_fun
    )
    t_final_val = jtu.tree_map(
        _perturb_to_tang, t_final_val, perturb_val, is_leaf=_is_none
    )
    return final_val, t_final_val



import jax.tree_util as jtu
from jaxtyping import PyTree

from .._tree import tree_at


def _inferences(pytree):
    is_leaf = lambda x: hasattr(x, "inference") and x is not pytree

    out = [pytree.inference] if hasattr(pytree, "inference") else []

    leaves = [x for x in jtu.tree_leaves(pytree, is_leaf=is_leaf) if is_leaf(x)]
    # Nodes with an inference flag might have sub-nodes with an inference flag.

    for x in leaves:
        out.extend(_inferences(x))
    return out


def inference_mode(pytree: PyTree, value: bool = True) -> PyTree:
    """Convenience function for setting all `inference` attributes.

    `inference` flags are used to toggle the behaviour of a number of the pre-built
    neural network layers, such as [`equinox.nn.Dropout`][] or
    [`equinox.nn.BatchNorm`][].

    !!! Example

        ```python
        class Model(eqx.Module):
            norm: eqx.nn.BatchNorm
            dropout: eqx.nn.Dropout
            linear: eqx.nn.Linear

            def __init__(self, key):
                key1, key2 = jax.random.split(key)
                self.norm = eqx.nn.BatchNorm(3, "batch", key=key1)
                self.dropout = eqx.nn.Dropout(0.4)
                self.linear = eqx.nn.Linear(3, 1, key=key2)

            def __call__(self, x, ctx, *, key):
                x, ctx = self.norm(x, ctx)
                x = self.dropout(x, key=key)
                x = self.linear(x)
                return x, ctx

        training_model = Model(jax.random.PRNGKey(0))
        inference_model = eqx.nn.inference_mode(training_model)
        training_model_again = eqx.nn.inference_mode(inference_model, value=False)
        ```

    **Arguments:**

    - `pytree`: the PyTree to modify.
    - `value`: the value to set all `inference` attributes to. Defaults to `True`, i.e.
        inference mode.

    **Returns:**

    A new PyTree with the same structure as the input PyTree with all `inference` flags
    set to `value`.
    """
    return tree_at(_inferences, pytree, replace_fn=lambda _: value)



"""Utilities for generating optimized flatten/unflatten functions for Module."""

import dataclasses
import textwrap
from enum import Enum
from typing import Any, Final, Literal, TypeVar

import jax.tree_util as jtu


# Names for generated functions
_FLAT_FUNC_NAME = "<generated_flatten_{0}>"
_FLAT_KEYS_NAME = "<generated_flatten_with_keys_{0}>"
_UNFLAT_NAME = "<generated_unflatten_{0}>"

# Fields of Module wrappers to always include
WRAPPER_FIELD_NAMES: Final = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)

# Indentation for generated code
_INDENT: Final = " " * 4


# Sentinel values for flattening/unflattening
class _Sentinel(Enum):
    """Sentinel values for flattening/unflattening."""

    MISSING = "MISSING"


MISSING = _Sentinel.MISSING

# Code template for flattening the wrapper fields
_GETTER = "get({k!r},MISSING)"  # get = obj.__dict__.get
_FLAT_WRAPPERS = ", ".join([_GETTER.format(k=k) for k in WRAPPER_FIELD_NAMES])

# Code template for flatten function
_FLAT_CODE = f'''
def {{func_name}}(obj: module_cls) -> {{return_annotation}}:
    """Generated {{func_name}} function for {{qualname}}.

    Dynamic fields: {{dynamic_fs}}
    Static fields: {{static_fs}}
    Wrapper fields: {WRAPPER_FIELD_NAMES}
    """
    get = obj.__dict__.get
    return (
        {{dynamic_vals}},
        (
            MISSING if len(obj.__dict__) == {{num_fields}} else ({_FLAT_WRAPPERS}),
            {{static_vals}}
        )
    )
'''

# Code template for setting wrapper fields during unflattening
_SET_WRAPPERS: Final = "\n".join(
    f"self.__dict__[{name!r}] = waux[{i}]" for i, name in enumerate(WRAPPER_FIELD_NAMES)
)

# Code template for setting dynamic fields during unflattening
_SET_DYNAMIC = """
if data[{i}] is not MISSING:
    object.__setattr__(self, {name!r}, data[{i}])
"""[1:-1]  # (trim leading and trailing newlines)

# Code template for setting static fields during unflattening
_SET_STATIC = """
if aux[{i}] is not MISSING:
    object.__setattr__(self, {name!r}, aux[{i}])
"""[1:-1]  # (trim leading and trailing newlines)

# Code template for unflatten function
_UNFLAT_FUNC = f'''
def unflatten(
    module_cls: type[T],
    aux: {{aux_type}},
    data: {{dynamic_type}},
) -> T:
    """Generated unflatten function for {{qualname}}.

    Dynamic fields: {{dynamic_fs}}
    Static fields: {{static_fs}}
    Wrapper fields: {WRAPPER_FIELD_NAMES}
    """
    self = object.__new__(module_cls)
    # Set fields directly by index
{{setters_dynamic}}
    if aux[0] is not MISSING:
        waux = aux[0]
{textwrap.indent(_SET_WRAPPERS, _INDENT * 2)}
{{setters_static}}
    return self
'''

_NS_BASE = {
    "object": object,
    "Any": Any,
    "tuple": tuple,
    "Literal": Literal,
    "MISSING": MISSING,
}


def _make_tuple_type(count: int, elt: str, /) -> str:
    """Generate a tuple type annotation string for a given count of elements."""
    return "tuple[()]" if count == 0 else f"tuple[{', '.join([elt] * count)}]"


def generate_flatten_functions(cls: type, fields: tuple[dataclasses.Field[Any], ...]):
    """Generate optimized flatten/unflatten functions for a specific field config."""
    # Separate dynamic and static fields
    _dynamic_fs, _static_fs = [], []
    for f in fields:
        if f.metadata.get("static", False):
            _static_fs.append(f.name)
        else:
            _dynamic_fs.append(f.name)
    dynamic_fs, static_fs = tuple(_dynamic_fs), tuple(_static_fs)
    n_dynamic_fs = len(dynamic_fs)
    n_static_fs = len(static_fs)

    # Extract the generated functions from respective namespaces to
    # set the proper module reference.
    module_name = getattr(cls, "__module__", cls.__name__)

    # -------------------------------------------
    # Generate flatten function

    # Directly access dynamic fields by name
    if n_dynamic_fs == 0:
        dynamic_vals = "()"
    else:
        dynamic_exprs = [_GETTER.format(k=k) for k in dynamic_fs]
        dynamic_vals = f"({', '.join(dynamic_exprs)},)"

    # For static fields, we need to store their values in aux data
    static_exprs = [_GETTER.format(k=k) for k in static_fs]
    static_vals = f"{', '.join(static_exprs)}"

    # Build return type annotation
    dynamic_type = _make_tuple_type(n_dynamic_fs, "Any")
    wrapper_type = "Literal[MISSING]|tuple[Any, ...]"
    static_type = f"tuple[{wrapper_type}, {', '.join(['Any'] * n_static_fs)}]"
    clsname = cls.__qualname__

    # Generate flatten function code
    flat_code = _FLAT_CODE.format(
        func_name="flatten",
        return_annotation=f"tuple[{dynamic_type}, {static_type}]",
        qualname=clsname,
        dynamic_fs=dynamic_fs,
        static_fs=static_fs,
        num_fields=n_dynamic_fs + n_static_fs,
        dynamic_vals=dynamic_vals,
        static_vals=static_vals,
    )

    # make flatten func
    flat_ns = _NS_BASE | {"module_cls": cls}
    exec(compile(flat_code, _FLAT_FUNC_NAME.format(clsname), "exec"), flat_ns)
    flat_fn = flat_ns["flatten"]
    flat_fn.__module__ = module_name
    object.__setattr__(flat_fn, "__source__", flat_code)

    # -------------------------------------------
    # Generate flatten_with_keys function

    # Generate flatten_with_keys values
    if n_dynamic_fs == 0:
        dynamic_key_vals = "()"
    else:
        key_exprs = [
            f"(jtu.GetAttrKey({k!r}), {_GETTER.format(k=k)})" for k in dynamic_fs
        ]
        dynamic_key_vals = f"({', '.join(key_exprs)},)"

    keys_dynamic_type = _make_tuple_type(n_dynamic_fs, "tuple[jtu.GetAttrKey, str]")

    # Generate flatten_with_keys function code
    flat_k_code = _FLAT_CODE.format(
        func_name="flatten_with_keys",
        return_annotation=f"tuple[{keys_dynamic_type}, {static_type}]",
        qualname=clsname,
        dynamic_fs=dynamic_fs,
        static_fs=static_fs,
        num_fields=n_dynamic_fs + n_static_fs,
        dynamic_vals=dynamic_key_vals,
        static_vals=static_vals,
    )

    # flatten with keys func
    flat_k_ns = _NS_BASE | {"jtu": jtu, "module_cls": cls}
    exec(compile(flat_k_code, _FLAT_KEYS_NAME.format(clsname), "exec"), flat_k_ns)
    flat_k_fn = flat_k_ns["flatten_with_keys"]
    flat_k_fn.__module__ = module_name
    object.__setattr__(flat_k_fn, "__source__", flat_k_code)

    # -------------------------------------------
    # Generate unflatten function - directly set fields by index
    # Extract types from flatten return type: tuple[dynamic_type, static_type]

    # Set dynamic fields by index
    unflat_dynamic = [
        _SET_DYNAMIC.format(i=i, name=k) for i, k in enumerate(dynamic_fs)
    ]
    # Set static fields by index. Offset by 1 for wrapper auxiliary field.
    unflat_aux = [
        _SET_STATIC.format(i=i, name=k) for i, k in enumerate(static_fs, start=1)
    ]

    # Generate unflatten function code
    unflat_code = _UNFLAT_FUNC.format(
        aux_type=static_type,
        dynamic_type=dynamic_type,
        qualname=clsname,
        dynamic_fs=dynamic_fs,
        static_fs=static_fs,
        setters_dynamic=textwrap.indent("\n".join(unflat_dynamic), _INDENT),
        setters_static=textwrap.indent("\n".join(unflat_aux), _INDENT),
    )

    # Namespace for unflatten function (takes module_cls as parameter)
    unflat_ns = _NS_BASE | {"type": type, "T": TypeVar("T")}
    # unflatten
    exec(compile(unflat_code, _UNFLAT_NAME.format(clsname), "exec"), unflat_ns)
    unflat_fn = unflat_ns["unflatten"]
    unflat_fn.__module__ = module_name
    object.__setattr__(unflat_fn, "__source__", unflat_code)

    # -------------------------------------------

    return flat_fn, flat_k_fn, unflat_fn



import warnings

import jax.lax as lax
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import Array, PRNGKeyArray

from .._module import Module
from ._misc import named_scope


class Dropout(Module):
    """Applies dropout.

    Note that this layer behaves differently during training and inference. During
    training then dropout is randomly applied; during inference this layer does nothing.
    Whether the model is in training or inference mode should be toggled using
    [`equinox.nn.inference_mode`][].
    """

    # Not static fields as it makes sense to want to modify them via equinox.tree_at.
    p: float
    inference: bool

    def __init__(
        self,
        p: float = 0.5,
        inference: bool = False,
        *,
        deterministic: bool | None = None,
    ):
        """**Arguments:**

        - `p`: The fraction of entries to set to zero. (On average.)
        - `inference`: Whether to actually apply dropout at all. If `True` then dropout
            is *not* applied. If `False` then dropout is applied. This may be toggled
            with [`equinox.nn.inference_mode`][] or overridden during
            [`equinox.nn.Dropout.__call__`][].
        - `deterministic`: Deprecated alternative to `inference`.
        """

        if deterministic is not None:
            inference = deterministic
            warnings.warn(
                "Dropout(deterministic=...) is deprecated "
                "in favour of Dropout(inference=...)"
            )
        self.p = p
        self.inference = inference

    # Backward compatibility
    @property
    def deterministic(self):
        return self.inference

    @named_scope("eqx.nn.Dropout")
    def __call__(
        self,
        x: Array,
        *,
        key: PRNGKeyArray | None = None,
        inference: bool | None = None,
        deterministic: bool | None = None,
    ) -> Array:
        """**Arguments:**

        - `x`: An any-dimensional JAX array to dropout.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for calculating
            which elements to dropout. (Keyword only argument.)
        - `inference`: As per [`equinox.nn.Dropout.__init__`][]. If `True` or
            `False` then it will take priority over `self.inference`. If `None`
            then the value from `self.inference` will be used.
        - `deterministic`: Deprecated alternative to `inference`.
        """

        if deterministic is not None:
            inference = deterministic
            warnings.warn(
                "Dropout()(deterministic=...) is deprecated "
                "in favour of Dropout()(inference=...)"
            )

        if inference is None:
            inference = self.inference
        if isinstance(self.p, (int, float)) and self.p == 0:
            inference = True
        if inference:
            return x
        elif key is None:
            raise RuntimeError(
                "Dropout requires a key when running in non-deterministic mode."
            )
        else:
            q = 1 - lax.stop_gradient(self.p)
            mask = jrandom.bernoulli(key, q, x.shape)
            return jnp.where(mask, x / q, 0)



import wadler_lindig as wl

from .._module import Module


def str2jax(msg: str):
    """Creates a JAXable object whose `str(...)` is the specified string."""

    class M(Module):
        def __pdoc__(self, **kwargs):
            del kwargs
            return wl.TextDoc(msg)

        def __repr__(self):
            return msg

    M.__name__ = M.__qualname__ = msg
    return M()



import types
from collections.abc import Callable
from typing import Any, Generic, TypeVar

import jax.tree_util as jtu
from jaxtyping import PyTreeDef

from ._field import field
from ._flatten import WRAPPER_FIELD_NAMES
from ._module import Module


# Not using `jax.tree_util.Partial` as it doesn't implement __eq__ very well. See #480.
class BoundMethod(Module):
    """Just like a normal Python bound method... except that this one is a PyTree!

    This stores `__self__` as a subnode.
    """

    __func__: types.FunctionType = field(static=True)
    __self__: Module

    def __post_init__(self):
        for field_name in WRAPPER_FIELD_NAMES:
            try:
                value = getattr(self.__func__, field_name)
            except AttributeError:
                pass
            else:
                setattr(self, field_name, value)

    def __call__(self, *args, **kwargs):
        __tracebackhide__ = True
        return self.__func__(self.__self__, *args, **kwargs)

    @property
    def __wrapped__(self):
        return self.__func__.__get__(self.__self__, type(self.__self__))  # pyright: ignore[reportAttributeAccessIssue]


_Return = TypeVar("_Return")


class Partial(Module, Generic[_Return]):
    """Like `functools.partial`, but treats the wrapped function, and partially-applied
    args and kwargs, as a PyTree.

    This is very much like `jax.tree_util.Partial`. The difference is that the JAX
    version requires that `func` be specifically a *function* -- and will silently
    misbehave if given any non-function callable, e.g. [`equinox.nn.MLP`][]. In contrast
    the Equinox version allows for arbitrary callables.
    """

    func: Callable[..., _Return]
    args: tuple[Any, ...]
    keywords: dict[str, Any]

    def __init__(self, func: Callable[..., _Return], /, *args: Any, **kwargs: Any):
        """**Arguments:**

        - `func`: the callable to partially apply.
        - `*args`: any positional arguments to apply.
        - `**kwargs`: any keyword arguments to apply.
        """
        self.func = func
        self.args = args
        self.keywords = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> _Return:
        """Call the wrapped `self.func`.

        **Arguments:**

        - `*args`: the arguments to apply. Passed after those arguments passed during
            `__init__`.
        - `**kwargs`: any keyword arguments to apply.

        **Returns:**

        The result of the wrapped function.
        """
        return self.func(*self.args, *args, **kwargs, **self.keywords)


_Value = TypeVar("_Value")


class Static(Module, Generic[_Value]):
    """Wraps a value into a `eqx.field(static=True)`.

    This is useful to treat something as just static metadata with respect to a JAX
    transformation; for example this is used to return non-arrays from a filtered
    transform.
    """

    _leaves: list[Any] = field(static=True)
    _treedef: PyTreeDef = field(static=True)  # pyright: ignore

    def __init__(self, value: _Value):
        # By flattening, we handle pytrees without `__eq__` methods.
        # When comparing static metadata for equality, this means we never actually
        # call `value.__eq__`.
        self._leaves, self._treedef = jtu.tree_flatten(value)

    @property
    def value(self) -> _Value:
        return jtu.tree_unflatten(self._treedef, self._leaves)



from collections.abc import Callable
from typing import (
    Literal,
)

import jax.nn as jnn
import jax.random as jrandom
import jax.tree_util as jtu
from jaxtyping import Array, PRNGKeyArray

from .._doc_utils import doc_repr
from .._filters import is_array
from .._misc import default_floating_dtype
from .._module import field, Module
from .._vmap_pmap import filter_vmap
from ._linear import Linear
from ._misc import named_scope


_identity = doc_repr(lambda x: x, "lambda x: x")
_relu = doc_repr(jnn.relu, "<function relu>")


class MLP(Module):
    """Standard Multi-Layer Perceptron; also known as a feed-forward network.

    !!! faq

        If you get a TypeError saying an object is not a valid JAX type, see the
            [FAQ](https://docs.kidger.site/equinox/faq/)."""

    layers: tuple[Linear, ...]
    activation: Callable
    final_activation: Callable
    use_bias: bool = field(static=True)
    use_final_bias: bool = field(static=True)
    in_size: int | Literal["scalar"] = field(static=True)
    out_size: int | Literal["scalar"] = field(static=True)
    width_size: int = field(static=True)
    depth: int = field(static=True)

    def __init__(
        self,
        in_size: int | Literal["scalar"],
        out_size: int | Literal["scalar"],
        width_size: int,
        depth: int,
        activation: Callable = _relu,
        final_activation: Callable = _identity,
        use_bias: bool = True,
        use_final_bias: bool = True,
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments**:

        - `in_size`: The input size. The input to the module should be a vector of
            shape `(in_features,)`
        - `out_size`: The output size. The output from the module will be a vector
            of shape `(out_features,)`.
        - `width_size`: The size of each hidden layer.
        - `depth`: The number of hidden layers, including the output layer.
            For example, `depth=2` results in an network with layers:
            [`Linear(in_size, width_size)`, `Linear(width_size, width_size)`,
            `Linear(width_size, out_size)`].
        - `activation`: The activation function after each hidden layer. Defaults to
            ReLU.
        - `final_activation`: The activation function after the output layer. Defaults
            to the identity.
        - `use_bias`: Whether to add on a bias to internal layers. Defaults
            to `True`.
        - `use_final_bias`: Whether to add on a bias to the final layer. Defaults
            to `True`.
        - `dtype`: The dtype to use for all the weights and biases in this MLP.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        Note that `in_size` also supports the string `"scalar"` as a special value.
        In this case the input to the module should be of shape `()`.

        Likewise `out_size` can also be a string `"scalar"`, in which case the
        output from the module will have shape `()`.
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        keys = jrandom.split(key, depth + 1)
        layers = []
        if depth == 0:
            layers.append(
                Linear(in_size, out_size, use_final_bias, dtype=dtype, key=keys[0])
            )
        else:
            layers.append(
                Linear(in_size, width_size, use_bias, dtype=dtype, key=keys[0])
            )
            for i in range(depth - 1):
                layers.append(
                    Linear(
                        width_size, width_size, use_bias, dtype=dtype, key=keys[i + 1]
                    )
                )
            layers.append(
                Linear(width_size, out_size, use_final_bias, dtype=dtype, key=keys[-1])
            )
        self.layers = tuple(layers)
        self.in_size = in_size
        self.out_size = out_size
        self.width_size = width_size
        self.depth = depth
        # In case `activation` or `final_activation` are learnt, then make a separate
        # copy of their weights for every neuron.
        self.activation = filter_vmap(
            filter_vmap(lambda: activation, axis_size=width_size), axis_size=depth
        )()
        if out_size == "scalar":
            self.final_activation = final_activation
        else:
            self.final_activation = filter_vmap(
                lambda: final_activation, axis_size=out_size
            )()
        self.use_bias = use_bias
        self.use_final_bias = use_final_bias

    @named_scope("eqx.nn.MLP")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: A JAX array with shape `(in_size,)`. (Or shape `()` if
            `in_size="scalar"`.)
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array with shape `(out_size,)`. (Or shape `()` if `out_size="scalar"`.)
        """
        for i, layer in enumerate(self.layers[:-1]):
            x = layer(x)
            layer_activation = jtu.tree_map(
                lambda x: x[i] if is_array(x) else x, self.activation
            )
            x = filter_vmap(lambda a, b: a(b))(layer_activation, x)
        x = self.layers[-1](x)
        if self.out_size == "scalar":
            x = self.final_activation(x)
        else:
            x = filter_vmap(lambda a, b: a(b))(self.final_activation, x)
        return x



import math
from collections.abc import Callable, Sequence

import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.random
from jaxtyping import Array, PRNGKeyArray

from .._module import field, Module
from ._misc import all_sequences, named_scope


class Pool(Module):
    """General N-dimensional downsampling over a sliding window."""

    init: int | float | Array
    operation: Callable[[Array, Array], Array]
    num_spatial_dims: int = field(static=True)
    kernel_size: tuple[int, ...] = field(static=True)
    stride: tuple[int, ...] = field(static=True)
    padding: tuple[tuple[int, int], ...] = field(static=True)
    use_ceil: bool = field(static=True)

    def __init__(
        self,
        init: int | float | Array,
        operation: Callable[[Array, Array], Array],
        num_spatial_dims: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `init`: The initial value for the reduction.
        - `operation`: The operation applied to the inputs of each window.
        - `num_spatial_dims`: The number of spatial dimensions.
        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.

        !!! info

            In order for `Pool` to be differentiable, `operation(init, x) == x` needs to
            be true for all finite `x`. For further details see
            [https://www.tensorflow.org/xla/operation_semantics#reducewindow](https://www.tensorflow.org/xla/operation_semantics#reducewindow)
            and [https://github.com/google/jax/issues/7718](https://github.com/google/jax/issues/7718).
        """  # noqa: E501

        self.operation = operation
        self.init = init
        self.num_spatial_dims = num_spatial_dims
        self.use_ceil = use_ceil

        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size,) * num_spatial_dims
        elif isinstance(kernel_size, Sequence):
            self.kernel_size = tuple(kernel_size)
        else:
            raise ValueError(
                "`kernel_size` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints."
            )

        if isinstance(stride, int):
            self.stride = (stride,) * num_spatial_dims
        elif isinstance(stride, Sequence):
            self.stride = tuple(stride)
        else:
            raise ValueError(
                "`stride` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints."
            )

        if isinstance(padding, int):
            self.padding = tuple((padding, padding) for _ in range(num_spatial_dims))
        elif isinstance(padding, Sequence) and len(padding) == num_spatial_dims:
            if all_sequences(padding):
                self.padding = tuple(padding)  # pyright: ignore
            else:
                self.padding = tuple((p, p) for p in padding)
        else:
            raise ValueError(
                "`padding` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints or tuples of length 2."
            )

    def _update_padding_for_ceil(self, input_shape):
        new_padding = []
        for input_size, (left_padding, right_padding), kernel_size, stride in zip(
            input_shape[1:], self.padding, self.kernel_size, self.stride
        ):
            if (input_size + left_padding + right_padding - kernel_size) % stride == 0:
                new_padding.append((left_padding, right_padding))
            else:
                new_padding.append((left_padding, right_padding + stride))
        return tuple(new_padding)

    def _check_is_padding_valid(self, padding):
        for (left_padding, right_padding), kernel_size in zip(
            padding, self.kernel_size
        ):
            if max(left_padding, right_padding) > kernel_size:
                raise RuntimeError(
                    "Paddings should be less than the size of the kernel. "
                    f"Padding {(left_padding, right_padding)} received for kernel size "
                    f"{kernel_size}."
                )

    @named_scope("eqx.nn.Pool")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, ..., dim_N)`, where `N = num_spatial_dims`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, ..., new_dim_N)`.
        """
        assert len(x.shape) == self.num_spatial_dims + 1, (
            f"Input should have {self.num_spatial_dims} spatial dimensions, "
            f"but input has shape {x.shape}"
        )

        if self.use_ceil:
            padding = self._update_padding_for_ceil(x.shape)
        else:
            padding = self.padding

        self._check_is_padding_valid(padding)

        x = jnp.moveaxis(x, 0, -1)
        x = jnp.expand_dims(x, axis=0)
        x = lax.reduce_window(
            x,
            self.init,
            self.operation,
            (1,) + self.kernel_size + (1,),
            (1,) + self.stride + (1,),
            ((0, 0),) + padding + ((0, 0),),
        )

        x = jnp.squeeze(x, axis=0)
        x = jnp.moveaxis(x, -1, 0)
        return x


class AvgPool1d(Pool):
    """One-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=1,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    @named_scope("eqx.nn.AvgPool1d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim)`.
        """

        return super().__call__(x) / math.prod(self.kernel_size)


class MaxPool1d(Pool):
    """One-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=1,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    # Redefined to get them in the right order in docs
    @named_scope("eqx.nn.MaxPool1d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim)`.
        """

        return super().__call__(x)


class AvgPool2d(Pool):
    """Two-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    @named_scope("eqx.nn.AvgPool2d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2)`.
        """

        return super().__call__(x) / math.prod(self.kernel_size)


class MaxPool2d(Pool):
    """Two-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    # Redefined to get them in the right order in docs
    @named_scope("eqx.nn.MaxPool2d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2)`.
        """

        return super().__call__(x)


class AvgPool3d(Pool):
    """Three-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=3,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    @named_scope("eqx.nn.AvgPool3d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, dim_2, dim_3)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2, new_dim_3)`.
        """

        return super().__call__(x) / math.prod(self.kernel_size)


class MaxPool3d(Pool):
    """Three-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        use_ceil: bool = False,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        - `use_ceil`: If `True`, then `ceil` is used to compute the final output
            shape instead of `floor`. For `ceil`, if required, extra padding is added.
            Defaults to `False`.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=3,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            use_ceil=use_ceil,
        )

    @named_scope("eqx.nn.MaxPool3d")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, dim_2, dim_3)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2, new_dim_3)`.
        """

        return super().__call__(x)


def _adaptive_pool1d(
    x: Array, target_size: int, operation: Callable[[Array], Array]
) -> Array:
    """**Arguments:**

    - `x`: The input. Should be a JAX array of shape `(dim,)`.
    - `target_size`: The shape of the output after the pooling operation
        `(target_size,)`.
    - `operation`: The pooling operation to be performed on the input array.

    **Returns:**

    A JAX array of shape `(1, target_shape)`.
    """
    dims = jnp.size(x)
    num_head_arrays = dims % target_size
    if num_head_arrays != 0:
        head_end_index = num_head_arrays * (dims // target_size + 1)
        head_op = jax.vmap(operation)(x[:head_end_index].reshape(num_head_arrays, -1))
        tail_op = jax.vmap(operation)(
            x[head_end_index:].reshape(-1, dims // target_size)
        )
        outputs = jnp.concatenate([head_op, tail_op])
    else:
        outputs = jax.vmap(operation)(
            jax.vmap(operation)(x.reshape(-1, dims // target_size))
        )
    return outputs


class AdaptivePool(Module):
    """General N dimensional adaptive downsampling to a target shape."""

    target_shape: Sequence[int] = field(static=True)
    operation: Callable[[Array], Array]

    def __init__(
        self,
        target_shape: int | Sequence[int],
        num_spatial_dims: int,
        operation: Callable,
    ):
        """**Arguments:**

        - `target_shape`: The target output shape.
        - `num_spatial_dims`: The number of spatial dimensions.
        - `operation`: The operation applied for downsample.
        """
        self.operation = operation
        if isinstance(target_shape, int):
            self.target_shape = (target_shape,) * num_spatial_dims
        elif (
            isinstance(target_shape, Sequence) and len(target_shape) == num_spatial_dims
        ):
            self.target_shape = target_shape
        else:
            raise ValueError(
                "`target_size` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints."
            )

    @named_scope("eqx.nn.AdaptivePool")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, dim_2, ... )`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels,) + target_shape`.
        """
        if x.ndim - 1 != len(self.target_shape):
            raise ValueError(
                f"Expected input with {len(self.target_shape)} dimensions, "
                f"received {x.ndim - 1} instead."
            )
        for i in range(1, x.ndim):
            op = jax.vmap(
                _adaptive_pool1d, (0, None, None), 0
            )  # batching over channels by default
            for j in range(1, x.ndim):
                if i == j:
                    continue
                op = jax.vmap(op, in_axes=(j, None, None), out_axes=j)
            x = op(x, self.target_shape[i - 1], self.operation)
        return x


class AdaptiveAvgPool1d(AdaptivePool):
    """Adaptive one-dimensional downsampling using average for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.mean)


class AdaptiveAvgPool2d(AdaptivePool):
    """Adaptive two-dimensional downsampling using average for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.mean)


class AdaptiveAvgPool3d(AdaptivePool):
    """Adaptive three-dimensional downsampling using average for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.mean)


class AdaptiveMaxPool1d(AdaptivePool):
    """Adaptive one-dimensional downsampling using maximum for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.max)


class AdaptiveMaxPool2d(AdaptivePool):
    """Adaptive two-dimensional downsampling using maximum for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.max)


class AdaptiveMaxPool3d(AdaptivePool):
    """Adaptive three-dimensional downsampling using maximum for the target shape."""

    def __init__(self, target_shape: int | Sequence[int]):
        """**Arguments:**

        - `target_shape`: The target output shape.
        """
        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.max)



"""Defines a new transformation, "finalisation", which replaces all custom primitives
with the results of their `impl` rule.

This is useful prior to e.g. ONNX export, to replace all custom primitives (e.g.
`equinox.internal.unvmap_any`) with the JAX operations that they wrap.

Most end users will find `finalise_fn` and `finalise_make_jaxpr` to be the most useful
operations provided by this module.

Library authors may wish to register their primitives with `primitive_finalisations`.

!!! warning

    Note that you should not perform any other JAX transformations -- e.g. `jax.vmap`,
    `jax.grad` etc. -- after finalisation. The result may be silently incorrect.
"""

import functools as ft
from collections.abc import Callable
from typing import Any, cast, Literal, overload

import jax
import jax.core
import jax.custom_derivatives
import jax.extend.core
import jax.tree_util as jtu
from jaxtyping import PyTree


def _safe_map(f, *args):
    args = list(map(list, args))
    n = len(args[0])
    for arg in args[1:]:
        assert len(arg) == n, f"length mismatch: {list(map(len, args))}"
    return list(map(f, *args))


def _maybe_finalise_jaxpr(val: Any):
    is_open_jaxpr = False
    if isinstance(val, jax.extend.core.Jaxpr):
        if len(val.constvars) == 0:
            is_open_jaxpr = True
            val = jax.extend.core.ClosedJaxpr(val, [])
        else:
            return val
    if isinstance(val, jax.extend.core.ClosedJaxpr):
        val = finalise_jaxpr(val)
    if is_open_jaxpr:
        val = val.jaxpr
    return val


def _finalise_jaxprs_in_params(params):
    new_params = {}
    for key, val in params.items():
        if type(val) is tuple:  # not isinstance to avoid breaking namedtuples
            val = tuple(_maybe_finalise_jaxpr(v) for v in val)
        else:
            val = _maybe_finalise_jaxpr(val)
        new_params[key] = val
    return new_params


def _default_finalisation(prim: jax.extend.core.Primitive, *args, **kwargs):
    return prim.bind(*args, **kwargs)


def _impl_finalisation(prim: jax.extend.core.Primitive, *args, **kwargs):
    return prim.impl(*args, **kwargs)


primitive_finalisations = {}


def register_impl_finalisation(prim: jax.extend.core.Primitive):
    primitive_finalisations[prim] = ft.partial(_impl_finalisation, prim)


def finalise_eval_jaxpr(jaxpr: jax.extend.core.Jaxpr, consts, *args):
    """As jax.core.eval_jaxpr, but finalises (typically by calling `impl` rather than
    `bind` for custom primitives).
    """

    def read(v: jax.core.Atom) -> Any:
        return v.val if isinstance(v, jax.extend.core.Literal) else env[v]

    def write(v: jax.extend.core.Var, val: Any) -> None:
        env[v] = val

    env: dict[jax.extend.core.Var, Any] = {}
    _safe_map(write, jaxpr.constvars, consts)
    _safe_map(write, jaxpr.invars, args)
    for eqn in jaxpr.eqns:
        params = _finalise_jaxprs_in_params(eqn.params)
        subfuns, bind_params = eqn.primitive.get_bind_params(params)
        try:
            call = primitive_finalisations[eqn.primitive]
        except KeyError:
            call = ft.partial(_default_finalisation, eqn.primitive)
        ans = call(*subfuns, *_safe_map(read, eqn.invars), **bind_params)
        if eqn.primitive.multiple_results:
            _safe_map(write, eqn.outvars, ans)
        else:
            write(eqn.outvars[0], ans)
    return _safe_map(read, jaxpr.outvars)


def finalise_jaxpr_as_fn(jaxpr: jax.extend.core.ClosedJaxpr):
    """As `jax.core.jaxpr_as_fn`, but the result is finalised."""
    return ft.partial(finalise_eval_jaxpr, jaxpr.jaxpr, jaxpr.consts)


def finalise_jaxpr(jaxpr: jax.extend.core.ClosedJaxpr) -> jax.extend.core.ClosedJaxpr:
    """A jaxpr-to-jaxpr transformation that performs finalisation."""
    fn = finalise_jaxpr_as_fn(jaxpr)
    args = [
        jax.ShapeDtypeStruct(x.aval.shape, x.aval.dtype) for x in jaxpr.jaxpr.invars
    ]
    return cast(jax.extend.core.ClosedJaxpr, jax.make_jaxpr(fn)(*args))


def finalise_fn(fn):
    """Wraps a function to perform finalisation. (In a manner similar to e.g. `jax.vmap`
    wrapping a function to perform batching.)
    """

    def _finalise_fn(*args):
        jaxpr, struct = jax.make_jaxpr(fn, return_shape=True)(*args)  # pyright: ignore
        flat_args = jtu.tree_leaves(args)
        out = finalise_eval_jaxpr(jaxpr.jaxpr, jaxpr.consts, *flat_args)
        treedef = jtu.tree_structure(struct)
        return jtu.tree_unflatten(treedef, out)

    return _finalise_fn


@overload
def finalise_make_jaxpr(  # pyright: ignore[reportOverlappingOverload]
    fn, *, return_shape: Literal[False] = False
) -> Callable[..., jax.extend.core.ClosedJaxpr]: ...


@overload
def finalise_make_jaxpr(
    fn, *, return_shape: Literal[True] = True
) -> Callable[
    ..., tuple[jax.extend.core.ClosedJaxpr, PyTree[jax.ShapeDtypeStruct]]
]: ...


@overload
def finalise_make_jaxpr(
    fn, *, return_shape: bool = False
) -> Callable[
    ...,
    (
        jax.extend.core.ClosedJaxpr
        | tuple[jax.extend.core.ClosedJaxpr, PyTree[jax.ShapeDtypeStruct]]
    ),
]: ...


def finalise_make_jaxpr(fn, *, return_shape: bool = False):
    """As `jax.make_jaxpr`, but finalises the result."""

    def _finalise_make_jaxpr(*args):
        jaxpr_struct = jax.make_jaxpr(fn, return_shape=return_shape)(  # pyright: ignore
            *args
        )
        if return_shape:
            jaxpr_struct = cast(tuple[jax.extend.core.ClosedJaxpr, Any], jaxpr_struct)
            jaxpr, struct = jaxpr_struct
            jaxpr = finalise_jaxpr(jaxpr)
            return jaxpr, struct
        else:
            jaxpr_struct = cast(jax.extend.core.ClosedJaxpr, jaxpr_struct)
            jaxpr = finalise_jaxpr(jaxpr_struct)
            return jaxpr

    return _finalise_make_jaxpr


# Register finalisation rules for Equinox's custom primitives.
from .._unvmap import unvmap_all_p, unvmap_any_p, unvmap_max_p
from ..debug._announce_transform import announce_jaxpr_p
from ._loop import maybe_set_p, select_if_vmap_p
from ._noinline import noinline_p
from ._nontraceable import nonbatchable_p, nondifferentiable_backward_p, nontraceable_p


for prim in (
    unvmap_all_p,
    unvmap_any_p,
    unvmap_max_p,
    nondifferentiable_backward_p,
    announce_jaxpr_p,
    select_if_vmap_p,
    maybe_set_p,
    noinline_p,
    nontraceable_p,
    nonbatchable_p,
):
    register_impl_finalisation(prim)


# To make this also useful as debugging tool, we also inline some calls.


def _jvp_call_p_finalisation(fun, jvp, *args, symbolic_zeros=None):
    return fun.call_wrapped(*args)


def _stop_gradient_finalisation(x):
    return x


primitive_finalisations[jax.custom_derivatives.custom_jvp_call_p] = (
    _jvp_call_p_finalisation
)
primitive_finalisations[jax.lax.stop_gradient_p] = _stop_gradient_finalisation



# Backward compatibility: expose via `equinox.internal`. Now available under `equinox`.
from .._compile_utils import (
    hashable_combine as hashable_combine,
    hashable_partition as hashable_partition,
)
from .._doc_utils import doc_remove_args as doc_remove_args, doc_repr as doc_repr

# Backward compatibility: expose via `equinox.internal`. Now available under `equinox`.
from .._enum import Enumeration as Enumeration

# Backward compatibility: expose via `equinox.internal`. Now available under `equinox`.
from .._errors import (
    assert_dce as assert_dce,
    branched_error_if as branched_error_if,
    error_if as error_if,
)
from .._eval_shape import cached_filter_eval_shape as cached_filter_eval_shape
from .._misc import left_broadcast_to as left_broadcast_to
from .._module import (
    AbstractClassVar as AbstractClassVar,
    AbstractVar as AbstractVar,
    Static as Static,
)
from .._pretty_print import TreeWLCustom as TreeWLCustom
from .._unvmap import (
    unvmap_all as unvmap_all,
    unvmap_all_p as unvmap_all_p,
    unvmap_any as unvmap_any,
    unvmap_any_p as unvmap_any_p,
    unvmap_max as unvmap_max,
    unvmap_max_p as unvmap_max_p,
)

# Backward compatibility: expose via `equinox.internal`. Now available under
# `equinox.debug`.
from ..debug import (
    announce_transform as announce_transform,
    backward_nan as debug_backward_nan,  # noqa: F401
    breakpoint_if as breakpoint_if,
    inspect_dce as inspect_dce,
    store_dce as store_dce,
)
from ..debug._announce_transform import announce_jaxpr_p as announce_jaxpr_p
from ._closure_to_pytree import closure_to_pytree as closure_to_pytree
from ._finalise_jaxpr import (
    finalise_eval_jaxpr as finalise_eval_jaxpr,
    finalise_fn as finalise_fn,
    finalise_jaxpr as finalise_jaxpr,
    finalise_jaxpr_as_fn as finalise_jaxpr_as_fn,
    finalise_make_jaxpr as finalise_make_jaxpr,
    primitive_finalisations as primitive_finalisations,
    register_impl_finalisation as register_impl_finalisation,
)
from ._getkey import GetKey as GetKey
from ._loop import (
    buffer_at_set as buffer_at_set,
    maybe_set_p as maybe_set_p,
    MaybeBuffer as MaybeBuffer,
    scan as scan,
    select_if_vmap_p as select_if_vmap_p,
    while_loop as while_loop,
)
from ._misc import (
    ContainerMeta as ContainerMeta,
    eval_empty as eval_empty,
    eval_zero as eval_zero,
    scan_trick as scan_trick,
)
from ._nextafter import nextafter as nextafter, prevbefore as prevbefore
from ._noinline import noinline as noinline, noinline_p as noinline_p
from ._nontraceable import (
    nonbatchable as nonbatchable,
    nonbatchable_p as nonbatchable_p,
    nondifferentiable as nondifferentiable,
    nondifferentiable_backward as nondifferentiable_backward,
    nondifferentiable_backward_p as nondifferentiable_backward_p,
    nontraceable as nontraceable,
    nontraceable_p as nontraceable_p,
)
from ._omega import ω as ω
from ._onnx import to_onnx as to_onnx
from ._primitive import (
    create_vprim as create_vprim,
    filter_primitive_batching as filter_primitive_batching,
    filter_primitive_bind as filter_primitive_bind,
    filter_primitive_def as filter_primitive_def,
    filter_primitive_jvp as filter_primitive_jvp,
    filter_primitive_transpose as filter_primitive_transpose,
    materialise_zeros as materialise_zeros,
)
from ._str2jax import str2jax as str2jax



"""Introduces `AbstractVar` and `AbstractClassVar` type annotations, which can be used
to mark abstract instance attributes and abstract class attributes.

Provides enhanced versions of `abc.ABCMeta` and `dataclasses.dataclass` that respect
this type annotation. (In practice the sole consumer of these is `equinox.Module`, which
is the public API. But the two pieces aren't directly related under-the-hood.)
"""

import abc
import dataclasses
import inspect
from typing import (
    Annotated,
    ClassVar,
    Generic,
    get_args,
    get_origin,
    TYPE_CHECKING,
    TypeAlias,
    TypeVar,
)
from typing_extensions import dataclass_transform


_T = TypeVar("_T")


if TYPE_CHECKING:
    AbstractVar: TypeAlias = Annotated[_T, "AbstractVar"]
    from typing import ClassVar as AbstractClassVar
else:

    class AbstractVar(Generic[_T]):
        """Used to mark an abstract instance attribute, along with its type. Used as:
        ```python
        class Foo(eqx.Module):
            attr: AbstractVar[bool]
        ```

        An `AbstractVar[T]` must be overridden by an attribute annotated with
        `AbstractVar[T]`, `AbstractClassVar[T]`, `ClassVar[T]`, `T`, or a property
        returning `T`.

        This makes `AbstractVar` useful when you just want to assert that you can access
        `self.attr` on a subclass, regardless of whether it's an instance attribute,
        class attribute, property, etc.

        Attempting to instantiate a module with an unoveridden `AbstractVar` will raise
        an error.

        !!! Example

            ```python
            import equinox as eqx

            class AbstractX(eqx.Module):
                attr1: int
                attr2: AbstractVar[bool]

            class ConcreteX(AbstractX):
                attr2: bool

            ConcreteX(attr1=1, attr2=True)
            ```

        !!! Info

            `AbstractVar` does not create a dataclass field. This affects the order of
            `__init__` arguments. E.g.
            ```python
            class AbstractX(Module):
                attr1: AbstractVar[bool]

            class ConcreteX(AbstractX):
                attr2: str
                attr1: bool
            ```
            should be called as `ConcreteX(attr2, attr1)`.
        """

    class AbstractClassVar(Generic[_T]):
        """Used to mark an abstract class attribute, along with its type. Used as:
        ```python
        class Foo(eqx.Module):
            attr: AbstractClassVar[bool]
        ```

        An `AbstractClassVar[T]` can be overridden by an attribute annotated with
        `AbstractClassVar[T]`, or `ClassVar[T]`. This makes `AbstractClassVar` useful
        when you want to assert that you can access `cls.attr` on a subclass.

        Attempting to instantiate a module with an unoveridden `AbstractClassVar` will
        raise an error.

        !!! Example

            ```python
            import equinox as eqx
            from typing import ClassVar

            class AbstractX(eqx.Module):
                attr1: int
                attr2: AbstractClassVar[bool]

            class ConcreteX(AbstractX):
                attr2: ClassVar[bool] = True

            ConcreteX(attr1=1)
            ```

        !!! Info

            `AbstractClassVar` does not create a dataclass field. This affects the order
            of `__init__` arguments. E.g.
            ```python
            class AbstractX(Module):
                attr1: AbstractClassVar[bool]

            class ConcreteX(AbstractX):
                attr2: str
                attr1: ClassVar[bool] = True
            ```
            should be called as `ConcreteX(attr2)`.

        ## Known issues

        Due to a Pyright bug
        ([#4965](https://github.com/microsoft/pyright/issues/4965)), this must be
        imported as:
        ```python
        if TYPE_CHECKING:
            from typing import ClassVar as AbstractClassVar
        else:
            from equinox import AbstractClassVar
        ```
        """


def _process_annotation(annotation):
    if isinstance(annotation, str):
        if annotation.startswith("AbstractVar[") or annotation.startswith(
            "AbstractClassVar["
        ):
            raise NotImplementedError(
                "Stringified abstract annotations are not supported"
            )
        else:
            is_abstract = False
            is_class = annotation.startswith("ClassVar[")
            return is_abstract, is_class
    else:
        if annotation in (AbstractVar, AbstractClassVar):
            raise TypeError(
                "Cannot use unsubscripted `AbstractVar` or `AbstractClassVar`."
            )
        elif get_origin(annotation) is AbstractVar:
            if len(get_args(annotation)) != 1:
                raise TypeError("`AbstractVar` can only have a single argument.")
            is_abstract = True
            is_class = False
        elif get_origin(annotation) is AbstractClassVar:
            if len(get_args(annotation)) != 1:
                raise TypeError("`AbstractClassVar` can only have a single argument.")
            is_abstract = True
            is_class = True
        elif get_origin(annotation) is ClassVar:
            is_abstract = False
            is_class = True
        else:
            is_abstract = False
            is_class = False
        return is_abstract, is_class


class BetterABCMeta(abc.ABCMeta):
    __abstractvars__: frozenset[str]
    __abstractclassvars__: frozenset[str]

    def register(cls, subclass):
        del subclass
        raise ValueError

    def __new__(mcs, name, bases, namespace, /, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        # We don't try and check that our AbstractVars and AbstractClassVars are
        # consistently annotated across `cls` and each element of `bases`. Python just
        # doesn't really provide any way of checking that two hints are compatible.
        # (Subscripted generics make this complicated!)

        abstract_vars = set()
        abstract_class_vars = set()
        for kls in reversed(cls.__mro__):
            ann = inspect.get_annotations(kls)
            for name, annotation in ann.items():
                is_abstract, is_class = _process_annotation(annotation)
                if is_abstract:
                    if is_class:
                        if name in kls.__dict__:
                            raise TypeError(
                                f"Abstract class attribute {name} cannot have value"
                            )
                        abstract_vars.discard(name)
                        abstract_class_vars.add(name)
                    else:
                        if name in kls.__dict__:
                            raise TypeError(
                                f"Abstract attribute {name} cannot have value"
                            )
                        # If it's already an abstract class var, then superfluous to
                        # also consider it an abstract var.
                        if name not in abstract_class_vars:
                            abstract_vars.add(name)
                else:
                    abstract_vars.discard(name)  # not conditional on `is_class`
                    if is_class:
                        abstract_class_vars.discard(name)
            for name in kls.__dict__.keys():
                abstract_vars.discard(name)
                abstract_class_vars.discard(name)
        cls.__abstractvars__ = frozenset(abstract_vars)  # pyright: ignore
        cls.__abstractclassvars__ = frozenset(abstract_class_vars)  # pyright: ignore
        return cls

    def __call__(cls, *args, **kwargs):
        __tracebackhide__ = True
        if len(cls.__abstractclassvars__) > 0:  # pyright: ignore
            abstract_class_vars = set(cls.__abstractclassvars__)  # pyright: ignore
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} with abstract class "
                f"attributes {abstract_class_vars}"
            )
        self = super().__call__(*args, **kwargs)
        if len(cls.__abstractvars__) > 0:  # pyright: ignore
            abstract_class_vars = set(cls.__abstractvars__)  # pyright: ignore
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} with abstract "
                f"attributes {abstract_class_vars}"
            )
        return self


@dataclass_transform()
def better_dataclass(**kwargs):
    def make_dataclass(cls):
        try:
            annotations = inspect.get_annotations(cls)
        except KeyError:
            cls = dataclasses.dataclass(**kwargs)(cls)
        else:
            new_annotations = dict(annotations)
            for name, annotation in annotations.items():
                is_abstract, _ = _process_annotation(annotation)
                if is_abstract:
                    new_annotations.pop(name)
            cls.__annotations__ = new_annotations
            cls = dataclasses.dataclass(**kwargs)(cls)
            cls.__annotations__ = annotations
        return cls

    return make_dataclass



# This is some mildly unpleasant code.
#
# Basically, Optax make the decision *not* to register their optimisers as PyTrees.
# This means that we often end up with spurious recompilation, just because a learning
# rate changed. That results in a new optimiser instance, which is just a function and
# is treated statically.
#
# So here we simply replace all function closures with pytrees, with each of their cell
# contents as their subnodes.

import inspect
import types
from typing import Any

import jax.tree_util as jtu

from .._module import Module


def _make_cell(val):
    fn = lambda: val
    return fn.__closure__[0]  # pyright: ignore


def _adjust_function_closure(fn, closure):
    out = types.FunctionType(
        code=fn.__code__,
        globals=fn.__globals__,
        name=fn.__name__,
        argdefs=fn.__defaults__,
        closure=closure,
    )
    out.__module__ = fn.__module__
    out.__qualname__ = fn.__qualname__
    out.__doc__ = fn.__doc__
    out.__annotations__.update(fn.__annotations__)
    if fn.__kwdefaults__ is not None:
        out.__kwdefaults__ = fn.__kwdefaults__.copy()
    return out


# Not a pytree.
# Used so that two different local functions, with different identities, can still
# compare equal. This is needed as these leaves are compared statically when
# filter-jit'ing.
class _FunctionWithEquality:
    def __init__(self, fn: types.FunctionType):
        self.fn = fn

    def information(self):
        try:
            # `fn` not defined in REPL.
            source = inspect.getsource(self.fn)
        except OSError:
            # `fn` defined in REPL. In practice this will lead to recompilations based
            # on function identity, but correctness >> speed.
            return self.fn
        else:
            return self.fn.__qualname__, self.fn.__module__, source

    def __hash__(self):
        return hash(self.information())

    def __eq__(self, other):
        return type(self) == type(other) and self.information() == other.information()


class _Closure(Module):
    fn: _FunctionWithEquality
    contents: tuple[Any, ...] | None

    def __init__(self, fn: types.FunctionType):
        self.fn = _FunctionWithEquality(fn)
        if fn.__closure__ is None:
            contents = None
        else:
            contents = tuple(
                closure_to_pytree(cell.cell_contents) for cell in fn.__closure__
            )
        self.contents = contents

    def __call__(self, *args, **kwargs):
        if self.contents is None:
            closure = None
        else:
            closure = tuple(_make_cell(contents) for contents in self.contents)
        fn = _adjust_function_closure(self.fn.fn, closure)
        return fn(*args, **kwargs)


def _fixup_closure(leaf):
    if isinstance(leaf, types.FunctionType):
        return _Closure(leaf)
    else:
        return leaf


def closure_to_pytree(tree):
    """Convert all function closures into pytree nodes.

    **Arguments:**

    - `tree`: Any pytree.

    **Returns:**

    A copy of `tree`, where all function closures have been replaced by a new object
    that is (a) callable like the original function, but (b) iterates over its
    `__closure__` as subnodes in the pytree.

    !!! Example

        ```python
        def some_fn():
            a = jnp.array(1.)

            @closure_to_pytree
            def f(x):
                return x + a

            print(jax.tree_util.tree_leaves(f))  # prints out `a`
        ```

    !!! Warning

        One annoying technical detail in the above example: we had to wrap the whole lot
        in a `some_fn`, so that we're in a local scope. Python treats functions at the
        global scope differently, and this conversion won't result in any global
        variable being treated as part of the pytree.

        In practice, the intended use case of this function is to fix Optax, which
        always uses local functions.
    """
    return jtu.tree_map(_fixup_closure, tree)



import typing_extensions as te
from collections.abc import Callable, Sequence
from typing import Any, TYPE_CHECKING, TypeVar

import jax
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import PRNGKeyArray


_T = TypeVar("_T", bound=Sequence)

if TYPE_CHECKING:
    # StrictTypeGuard is a pyright-specific extension that performs type narrowing in
    # the `else` branch as well:
    # https://github.com/microsoft/pyright/issues/3450
    def all_sequences(
        x: Sequence[Any] | Sequence[_T],
    ) -> "te.TypeIs[Sequence[_T]]": ...

    _S = TypeVar("_S")

    def named_scope(name: str) -> Callable[[_S], _S]: ...

else:
    # beartype doesn't like StrictTypeGuard
    def all_sequences(x: Sequence[Any] | Sequence[_T]) -> bool:
        return all(isinstance(xi, Sequence) for xi in x)

    named_scope = jax.named_scope


def default_init(
    key: PRNGKeyArray, shape: tuple[int, ...], dtype: Any, lim: float
) -> jax.Array:
    if jnp.issubdtype(dtype, jnp.complexfloating):
        real_dtype = jnp.finfo(dtype).dtype
        rkey, ikey = jrandom.split(key, 2)
        real = jrandom.uniform(rkey, shape, real_dtype, minval=-lim, maxval=lim)
        imag = jrandom.uniform(ikey, shape, real_dtype, minval=-lim, maxval=lim)
        return real.astype(dtype) + 1j * imag.astype(dtype)
    else:
        return jrandom.uniform(key, shape, dtype, minval=-lim, maxval=lim)



import dataclasses
import warnings
from collections.abc import Callable
from typing import Any


def field(
    *,
    converter: Callable[[Any], Any] | None = None,
    static: bool = False,
    **kwargs: Any,
) -> Any:
    """Equinox supports extra functionality on top of the default dataclasses.

    **Arguments:**

    - `converter`: a function to call on this field when the model is initialised. For
        example, `field(converter=jax.numpy.asarray)` to convert
        `bool`/`int`/`float`/`complex` values to JAX arrays. This is ran after the
        `__init__` method (i.e. when using a user-provided `__init__`), and after
        `__post_init__` (i.e. when using the default dataclass initialisation).
        If `converter` is `None`, then no converter is registered.
    - `static`: whether the field should not interact with any JAX transform at all (by
        making it part of the PyTree structure rather than a leaf).
    - `**kwargs`: All other keyword arguments are passed on to `dataclass.field`.

    !!! example "Example for `converter`"

        ```python
        class MyModule(eqx.Module):
            foo: Array = eqx.field(converter=jax.numpy.asarray)

        mymodule = MyModule(1.0)
        assert isinstance(mymodule.foo, jax.Array)
        ```

    !!! example "Example for `static`"

        ```python
        class MyModule(eqx.Module):
            normal_field: int
            static_field: int = eqx.field(static=True)

        mymodule = MyModule("normal", "static")
        leaves, treedef = jax.tree_util.tree_flatten(mymodule)
        assert leaves == ["normal"]
        assert "static" in str(treedef)
        ```

    `static=True` means that this field is not a node of the PyTree, so it does not
    interact with any JAX transforms, like JIT or grad. This means that it is usually a
    bug to make JAX arrays be static fields. `static=True` should very rarely be used.
    It is preferred to just filter out each field with `eqx.partition` whenever you need
    to select only some fields.
    """
    try:
        metadata = dict(kwargs.pop("metadata"))  # safety copy
    except KeyError:
        metadata = {}
    if "converter" in metadata:
        raise ValueError("Cannot use metadata with `converter` already set.")
    if "static" in metadata:
        raise ValueError("Cannot use metadata with `static` already set.")
    # We don't just use `lambda x: x` as the default, so that this works:
    # ```
    # class Abstract(eqx.Module):
    #     x: int = eqx.field()
    #
    # class Concrete(Abstract):
    #    @property
    #    def x(self):
    #        pass
    # ```
    # otherwise we try to call the default converter on a property without a setter,
    # and an error is raised.
    # Oddities like the above are to be discouraged, of course, but in particular
    # `field(init=False)` was sometimes used to denote an abstract field (prior to the
    # introduction of `AbstractVar`), so we do want to support this.
    if converter is not None:
        metadata["converter"] = converter
    if static:
        metadata["static"] = True
    return dataclasses.field(metadata=metadata, **kwargs)


def static_field(**kwargs):
    warnings.warn(
        "`equinox.static_field` is deprecated in favour of "
        "`equinox.field(static=True)`",
        stacklevel=2,
    )
    return field(**kwargs, static=True)



import functools as ft
from collections.abc import Callable
from typing import Generic, TypeVar

import jax
import jax.numpy as jnp
from jaxtyping import Array, PRNGKeyArray, Scalar

from .._module import field, Module
from .._tree import tree_at
from ._misc import named_scope


_Layer = TypeVar("_Layer")


def _norm_except_axis(v: Array, norm: Callable[[Array], Scalar], axis: int | None):
    if axis is None:
        return norm(v)
    else:
        return jax.vmap(norm, in_axes=axis, out_axes=axis)(v)


class WeightNorm(Module, Generic[_Layer]):
    r"""Applies weight normalisation to a given parameter.

    Given the 2D weight matrix
    $W = (W_{ij})_{ij} \in \mathbb{R}^{\text{out}} \times \mathbb{R}^{\text{in}}$ of a
    linear layer, then it replaces it with the following reparameterisation:

    $g \frac{v_{ij}}{\lVert v_{i\, \cdot} \rVert} \in \mathbb{R}^{\text{out}} \times \mathbb{R}^{\text{in}}$

    where $v_{ij}$ is initialised as $W_{ij}$, and $g$ is initialised as
    $\lVert v_{i\, \cdot} \rVert = \sum_j {v_{ij}}^2$.

    Overall, the direction ($v$) and the magnitude ($g$) of the output of each neuron
    are treated separately.

    Given n-dimensional weight matrices $W$ (in convolutional layers), then the
    normalisation is analogusly instead computed over every axis except the first.

    ??? cite

        [Weight Normalisation](https://arxiv.org/abs/1602.07868)

        ```bibtex
        @article{DBLP:journals/corr/SalimansK16,
        author       = {Tim Salimans and
                        Diederik P. Kingma},
        title        = {Weight Normalisation: {A} Simple
                        Reparameterization to Accelerate
                        Training of Deep Neural Networks},
        journal      = {CoRR},
        volume       = {abs/1602.07868},
        year         = {2016},
        url          = {http://arxiv.org/abs/1602.07868},
        eprinttype   = {arXiv},
        eprint       = {1602.07868},
        timestamp    = {Mon, 13 Aug 2018 16:47:07 +0200},
        biburl       = {https://dblp.org/rec/journals/corr/SalimansK16.bib},
        bibsource    = {dblp computer science bibliography, https://dblp.org}
        }
        ```
    """  # noqa: E501

    layer: _Layer
    g: Array
    weight_name: str = field(static=True)
    axis: int | None = field(static=True)
    _norm: Callable[[Array], Scalar]

    def __init__(
        self,
        layer: _Layer,
        weight_name: str = "weight",
        axis: int | None = 0,
    ):
        """**Arguments:**

        - `layer`: The layer to wrap. Usually a [`equinox.nn.Linear`][] or
            a convolutional layer (e.g. [`equinox.nn.Conv2d`][]).
        - `weight_name`: The name of the layer's parameter (a JAX array) to apply
            weight normalisation to.
        - `axis`: The norm is computed across every axis except this one.
            If `None`, compute across every axis.
        """
        self.layer = layer
        self.weight_name = weight_name
        self.axis = axis

        self._norm = ft.partial(
            _norm_except_axis,
            norm=ft.partial(jnp.linalg.norm, keepdims=True),
            axis=axis,
        )
        self.g = self._norm(getattr(layer, weight_name))

    @named_scope("eqx.nn.WeightNorm")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: A JAX Array.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.

        **Returns:**

        - The JAX array from calling `self.layer(x)` (with weight normalisation
        applied).
        """

        v = getattr(self.layer, self.weight_name)
        weight = v * self.g / self._norm(v)
        layer = tree_at(lambda l: getattr(l, self.weight_name), self.layer, weight)
        return layer(x)



from typing import Generic, TypeVar

import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Float, PRNGKeyArray

from .._module import field
from .._tree import tree_at
from ._misc import named_scope
from ._sequential import StatefulLayer
from ._stateful import State, StateIndex


def _power_iteration(weight, u, v, eps):
    u = weight @ v
    u_norm = jnp.sqrt(jnp.sum(u**2))
    u = u / jnp.maximum(eps, u_norm)

    v = weight.T @ u
    v_norm = jnp.sqrt(jnp.sum(v**2))
    v = v / jnp.maximum(eps, v_norm)

    return u, v


_Layer = TypeVar("_Layer")


class SpectralNorm(StatefulLayer, Generic[_Layer]):
    """Applies spectral normalisation to a given parameter.

    Given a weight matrix $W$, and letting $σ(W)$ denote (an approximation to) its
    largest singular value, then this computes $W/σ(W)$.

    The approximation $σ(W)$ is computed using
    [power iterations](https://en.wikipedia.org/wiki/Power_iteration)
    which are updated (as a side-effect) every time $W/σ(W)$ is computed.

    Spectral normalisation is particularly commonly used when training generative
    adversarial networks; see
    [Spectral Normalization for Generative Adversarial Networks](https://arxiv.org/abs/1802.05957)
    for more details and motivation.

    !!! example

        See [this example](../../examples/stateful.ipynb) for example usage.

    Note that this layer behaves differently during training and inference. During
    training then power iterations are updated; during inference they are fixed.
    Whether the model is in training or inference mode should be toggled using
    [`equinox.nn.inference_mode`][].
    """  # noqa: E501

    layer: _Layer
    weight_name: str = field(static=True)
    uv_index: StateIndex[tuple[Float[Array, " u_size"], Float[Array, " v_size"]]]
    num_power_iterations: int = field(static=True)
    eps: float = field(static=True)
    inference: bool

    def __init__(
        self,
        layer: _Layer,
        weight_name: str,
        num_power_iterations: int = 1,
        eps: float = 1e-12,
        inference: bool = False,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments:**

        - `layer`: The layer to wrap. Usually a [`equinox.nn.Linear`][] or
            a convolutional layer (e.g. [`equinox.nn.Conv2d`][]).
        - `weight_name`: The name of the layer's parameter (a JAX array) to apply
            spectral normalisation to.
        - `num_power_iterations`: The number of power iterations to apply every time
            the array is accessed.
        - `eps`: Epsilon for numerical stability when calculating norms.
        - `inference`: Whether this is in inference mode, at which time no power
            iterations are performed.  This may be toggled with
            [`equinox.nn.inference_mode`][].
        - `key`: A `jax.random.PRNGKey` used to provide randomness for initialisation.
            (Keyword only argument.)


        !!! info

            The `dtype` of the weight array of the `layer` input is applied to all
            parameters in this layer.

        """
        self.layer = layer
        self.weight_name = weight_name
        self.num_power_iterations = num_power_iterations
        self.eps = eps
        self.inference = inference

        weight = getattr(layer, weight_name)
        if weight.ndim < 2:
            raise ValueError("`weight` must be at least two-dimensional")
        weight = jnp.reshape(weight, (weight.shape[0], -1))
        dtype = weight.dtype
        u_len, v_len = weight.shape
        ukey, vkey = jr.split(key)
        u0 = jr.normal(ukey, (u_len,), dtype=dtype)
        v0 = jr.normal(vkey, (v_len,), dtype=dtype)
        for _ in range(15):
            u0, v0 = _power_iteration(weight, u0, v0, eps)
        self.uv_index = StateIndex((u0, v0))

    @named_scope("eqx.nn.SpectralNorm")
    def __call__(
        self,
        x: Array,
        state: State,
        *,
        key: PRNGKeyArray | None = None,
        inference: bool | None = None,
    ) -> tuple[Array, State]:
        """**Arguments:**

        - `x`: A JAX array.
        - `state`: An [`equinox.nn.State`][] object (which is used to store the
            power iterations).
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)
        - `inference`: As per [`equinox.nn.SpectralNorm.__init__`][]. If
            `True` or `False` then it will take priority over `self.inference`. If
            `None` then the value from `self.inference` will be used.

        **Returns:**

        A 2-tuple of:

        - The JAX array from calling `self.layer(x)` (with spectral normalisation
            applied).
        - An updated context object (storing the updated power iterations).
        """

        u, v = state.get(self.uv_index)
        weight = getattr(self.layer, self.weight_name)
        weight_shape = weight.shape
        weight = jnp.reshape(weight, (weight.shape[0], -1))
        if inference is None:
            inference = self.inference
        if not inference:
            stop_weight = lax.stop_gradient(weight)
            for _ in range(self.num_power_iterations):
                u, v = _power_iteration(stop_weight, u, v, self.eps)
            state = state.set(self.uv_index, (u, v))
        σ = jnp.einsum("i,ij,j->", u, weight, v)
        σ_weight = jnp.reshape(weight / σ, weight_shape)
        layer = tree_at(lambda l: getattr(l, self.weight_name), self.layer, σ_weight)
        out = layer(x)
        return out, state



from ._better_abstract import (
    AbstractClassVar as AbstractClassVar,
    AbstractVar as AbstractVar,
)
from ._field import (
    field as field,
    static_field as static_field,  # deprecated
)
from ._module import (
    Module as Module,
    module_update_wrapper as module_update_wrapper,
    StrictConfig as StrictConfig,  # deprecated
)
from ._prebuilt import Partial as Partial, Static as Static



import jax
import jax.numpy as jnp
from jaxtyping import Array


@jax.custom_jvp
def nextafter(x: Array) -> Array:
    """Returns the floating point number after `x`."""
    y = jnp.nextafter(x, jnp.inf)
    # Flush denormal to normal.
    # Our use for these is to handle jumps in the vector field. Typically that means
    # there will be an "if x > cond" condition somewhere. However JAX uses DAZ
    # (denormals-are-zero), which will cause this check to fail near zero:
    # `jnp.nextafter(0, jnp.inf) > 0` gives `False`.
    return jnp.where(x == 0, jnp.finfo(x.dtype).tiny, y)  # pyright: ignore


@nextafter.defjvp
def nextafter_jvp(primals, tangents):
    (x,) = primals
    (tx,) = tangents
    return nextafter(x), tx


@jax.custom_jvp
def prevbefore(x: Array) -> Array:
    """Returns the floating point number before `x`."""
    y = jnp.nextafter(x, -jnp.inf)
    return jnp.where(x == 0, -jnp.finfo(x.dtype).tiny, y)  # pyright: ignore


@prevbefore.defjvp
def prevbefore_jvp(primals, tangents):
    (x,) = primals
    (tx,) = tangents
    return prevbefore(x), tx



from collections.abc import Callable
from typing import Any
from typing_extensions import ParamSpec

import jax
import jax._src.traceback_util as traceback_util
import jax.extend.core
import jax.tree_util as jtu
from jaxtyping import PyTree

from ._filters import combine, is_array, partition
from ._module import Module, module_update_wrapper, Static


traceback_util.register_exclusion(__file__)


_P = ParamSpec("_P")


def _is_struct(x):
    return is_array(x) or isinstance(x, jax.ShapeDtypeStruct)


class _MakeJaxpr(Module):
    fn: Callable

    @property
    def __wrapped__(self):
        return self.fn

    def __call__(self, *args, **kwargs):
        dynamic, static = partition((args, kwargs), _is_struct)
        dynamic_flat, dynamic_treedef = jtu.tree_flatten(dynamic)

        def _fn(*_dynamic_flat):
            _dynamic = jtu.tree_unflatten(dynamic_treedef, _dynamic_flat)
            _args, _kwargs = combine(_dynamic, static)
            _out = self.fn(*_args, **_kwargs)
            _out_dynamic, _out_static = partition(_out, is_array)
            return _out_dynamic, Static(_out_static)

        jaxpr, out_struct = jax.make_jaxpr(_fn, return_shape=True)(*dynamic_flat)
        dynamic_out_struct, static_out = out_struct
        static_out = static_out.value
        return jaxpr, dynamic_out_struct, static_out


def filter_make_jaxpr(
    fun: Callable[_P, Any],
) -> Callable[
    _P, tuple[jax.extend.core.ClosedJaxpr, PyTree[jax.ShapeDtypeStruct], PyTree[Any]]
]:
    """As `jax.make_jaxpr`, but accepts arbitrary PyTrees as input and output.

    **Arguments:**

    - `fun`: The function `fun(*arg, **kwargs)` whose jaxpr is to be computed. Its
        positional and keyword arguments may be anything, as can its return value.

    **Returns:**

    A wrapped version of `fun`, that when applied to example arguments
    `*args, **kwargs`, will return a 3-tuple of:

    - A `ClosedJaxpr` representing the evaluation of that function on those arguments.
    - A `PyTree[jax.ShapeDtypeStruct]` representing the output shape and dtype of the
        result.
    - A `PyTree[Any]` representing any non-array outputs from `fun`.

    The example arguments to be traced may be anything with `.shape` and `.dtype`
    fields (typically JAX arrays, NumPy arrays, of `jax.ShapeDtypeStruct`s). All
    other arguments are treated statically. In particular, Python builtins (`bool`,
    `int`, `float`, `complex`) are treated as static inputs; wrap them in JAX/NumPy
    arrays if you would like them to be traced.
    """
    return module_update_wrapper(_MakeJaxpr(fun))



from ._activations import PReLU as PReLU
from ._attention import MultiheadAttention as MultiheadAttention
from ._batch_norm import BatchNorm as BatchNorm
from ._conv import (
    Conv as Conv,
    Conv1d as Conv1d,
    Conv2d as Conv2d,
    Conv3d as Conv3d,
    ConvTranspose as ConvTranspose,
    ConvTranspose1d as ConvTranspose1d,
    ConvTranspose2d as ConvTranspose2d,
    ConvTranspose3d as ConvTranspose3d,
)
from ._dropout import Dropout as Dropout
from ._embedding import (
    Embedding as Embedding,
    RotaryPositionalEmbedding as RotaryPositionalEmbedding,
)
from ._inference import inference_mode as inference_mode
from ._linear import Identity as Identity, Linear as Linear
from ._mlp import MLP as MLP
from ._normalisation import (
    GroupNorm as GroupNorm,
    LayerNorm as LayerNorm,
    RMSNorm as RMSNorm,
)
from ._pool import (
    AdaptiveAvgPool1d as AdaptiveAvgPool1d,
    AdaptiveAvgPool2d as AdaptiveAvgPool2d,
    AdaptiveAvgPool3d as AdaptiveAvgPool3d,
    AdaptiveMaxPool1d as AdaptiveMaxPool1d,
    AdaptiveMaxPool2d as AdaptiveMaxPool2d,
    AdaptiveMaxPool3d as AdaptiveMaxPool3d,
    AdaptivePool as AdaptivePool,
    AvgPool1d as AvgPool1d,
    AvgPool2d as AvgPool2d,
    AvgPool3d as AvgPool3d,
    MaxPool1d as MaxPool1d,
    MaxPool2d as MaxPool2d,
    MaxPool3d as MaxPool3d,
    Pool as Pool,
)
from ._rnn import GRUCell as GRUCell, LSTMCell as LSTMCell
from ._sequential import (
    Lambda as Lambda,
    Sequential as Sequential,
    StatefulLayer as StatefulLayer,
)
from ._shared import Shared as Shared
from ._spectral_norm import SpectralNorm as SpectralNorm
from ._stateful import (
    delete_init_state as delete_init_state,
    make_with_state as make_with_state,
    State as State,
    StateIndex as StateIndex,
)
from ._weight_norm import WeightNorm as WeightNorm



import dataclasses
from collections.abc import Callable
from typing import Any

import jax
import jax.tree_util as jtu
import wadler_lindig as wl


@dataclasses.dataclass
class _Partial:
    func: Callable
    args: tuple[Any, ...]
    keywords: dict[str, Any]


_Partial.__name__ = jtu.Partial.__name__
_Partial.__qualname__ = jtu.Partial.__qualname__
_Partial.__module__ = jtu.Partial.__module__


def _false(_):
    return False


@dataclasses.dataclass(slots=True)
class TreeWLCustom:
    """Wadler-Lindig customisation for PyTree pretty-printing.

    **Arguments:**
    - `truncate_leaf`: A function `Any -> bool`. Applied to all nodes in the
        PyTree; all truthy nodes will be truncated to just "<class_name>(...)".
    - `short_arrays`: If `True`, arrays are abbreviated to their shape and
      dtype (with weak dtype info prepended to the dtype name).
    - `struct_as_array`: If `True`, JAX ShapeDtypeStructs are abbreviated to
      their shape and dtype.

    """

    truncate_leaf: Callable[[Any], bool]

    short_arrays: bool

    struct_as_array: bool

    def __call__(self, obj: Any) -> wl.AbstractDoc | None:
        if self.truncate_leaf(obj):
            return wl.TextDoc(f"{type(obj).__name__}(...)")

        if self.short_arrays:
            if isinstance(obj, jax.Array) or (
                self.struct_as_array and isinstance(obj, jax.ShapeDtypeStruct)
            ):
                dtype = obj.dtype.name
                # Added in JAX 0.4.32 to `ShapeDtypeStruct`
                if getattr(obj, "weak_type", False):
                    dtype = f"weak_{dtype}"
                return wl.array_summary(obj.shape, dtype, kind=None)

        if isinstance(obj, (jax.custom_jvp, jax.custom_vjp)):
            return wl.pdoc(obj.__wrapped__)

        if isinstance(obj, jtu.Partial):
            obj = _Partial(obj.func, obj.args, obj.keywords)
            return wl.pdoc(obj)


def tree_pformat(
    pytree: Any,
    *,
    width: int = 80,
    indent: int = 2,
    short_arrays: bool = True,
    struct_as_array: bool = False,
    truncate_leaf: Callable[[Any], bool] = _false,
) -> str:
    """Pretty-formats a PyTree as a string, whilst abbreviating JAX arrays.

    (This is the function used in `__repr__` of [`equinox.Module`][].)

    As [`equinox.tree_pprint`][], but returns the string instead of printing it.
    """
    return wl.pformat(
        pytree,
        width=width,
        indent=indent,
        short_arrays=short_arrays,
        custom=TreeWLCustom(
            truncate_leaf=truncate_leaf,
            short_arrays=short_arrays,
            struct_as_array=struct_as_array,
        ),
    )


def tree_pprint(
    pytree: Any,
    *,
    width: int = 80,
    indent: int = 2,
    short_arrays: bool = True,
    struct_as_array: bool = False,
    truncate_leaf: Callable[[Any], bool] = _false,
) -> None:
    """Pretty-prints a PyTree as a string, whilst abbreviating JAX arrays.

    All JAX arrays in the PyTree are condensed down to a short string representation
    of their dtype and shape.

    !!! example

        A 32-bit floating-point JAX array of shape `(3, 4)` is printed as `f32[3,4]`.

    **Arguments:**

    - `pytree`: The PyTree to pretty-print.
    - `width`: The desired maximum number of characters per line of output. If a
        structure cannot be formatted within this constraint then a best effort will
        be made.
    - `indent`: The amount of indentation each nesting level.
    - `short_arrays`: Toggles the abbreviation of JAX arrays.
    - `struct_as_array`: Whether to treat `jax.ShapeDtypeStruct`s as arrays.
    - `truncate_leaf`: A function `Any -> bool`. Applied to all nodes in the PyTree;
        all truthy nodes will be truncated to just `f"{type(node).__name__}(...)"`.

    **Returns:**

    Nothing. (The result is printed to stdout instead.)
    """
    print(
        tree_pformat(
            pytree,
            width=width,
            indent=indent,
            short_arrays=short_arrays,
            struct_as_array=struct_as_array,
            truncate_leaf=truncate_leaf,
        )
    )



import warnings
from collections.abc import Hashable, Sequence
from typing import Literal

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Int, PRNGKeyArray

from .._misc import default_floating_dtype
from .._module import field
from ._misc import named_scope
from ._sequential import StatefulLayer
from ._stateful import State, StateIndex


class BatchNorm(StatefulLayer):
    r"""Computes a mean and standard deviation over the batch and spatial
    dimensions of an array, and uses these to normalise the whole array. Optionally
    applies a channelwise affine transformation afterwards.

    Given an input array $x = [x_1, ... x_C]$ with $C$ channels, this layer computes

    $$\frac{x_i - \mathbb{E}[x_i]}{\sqrt{\text{Var}[x_i] + \varepsilon}} * \gamma_i + \beta_i$$

    for all $i$. Here $*$ denotes elementwise multiplication and $\gamma$, $\beta$ have
    shape $(C,)$ if `channelwise_affine=True` and $\gamma = 1$, $\beta = 0$ if
    `channelwise_affine=False`. Expectations are computed over all spatial dimensions
    *and* over the batch dimension, and updated batch-by-batch according to `momentum`.

    !!! example

        See [this example](../../examples/stateful.ipynb) for example usage.

    !!! warning

        This layer must be used inside of a `vmap` or `pmap` with a matching
        `axis_name`. (Not doing so will raise a `NameError`.)

    Note that this layer behaves differently during training and inference. During
    training then statistics are computed using the input data, and the running
    statistics updated. During inference then just the running statistics are used.
    Whether the model is in training or inference mode should be toggled using
    [`equinox.nn.inference_mode`][].

    With `mode = "batch"` during training the batch mean and variance are used
    for normalization.  For inference the exponential running mean and unbiased
    variance are used for normalization. This is in line with how other machine
    learning packages (e.g. PyTorch, flax, haiku) implement batch norm.

    With `mode = "ema"` exponential running means and variances are kept.  During
    training the batch statistics are used to fill in the running statistics until
    they are populated.  During inference the running statistics are used for
    normalization.

    ??? cite

        [Batch Normalization: Accelerating Deep Network Training by Reducing
         Internal Covariate Shift](https://arxiv.org/abs/1502.03167)

        ```bibtex
        @article{DBLP:journals/corr/IoffeS15,
            author = {Sergey Ioffe and Christian Szegedy},
            title = {Batch Normalization: Accelerating Deep Network Training by Reducing
                     Internal Covariate Shift},
            journal = {CoRR},
            volume = {abs/1502.03167},
            year = {2015},
            url = {http://arxiv.org/abs/1502.03167},
            eprinttype = {arXiv},
            eprint = {1502.03167},
            timestamp = {Mon, 13 Aug 2018 16:47:06 +0200},
            biburl = {https://dblp.org/rec/journals/corr/IoffeS15.bib},
            bibsource = {dblp computer science bibliography, https://dblp.org}
        }
        ```
    """  # noqa: E501

    weight: Float[Array, "input_size"] | None
    bias: Float[Array, "input_size"] | None
    ema_first_time_index: None | StateIndex[Bool[Array, ""]]
    ema_state_index: (
        None | StateIndex[tuple[Float[Array, "input_size"], Float[Array, "input_size"]]]
    )
    batch_counter: None | StateIndex[Int[Array, ""]]
    batch_state_index: (
        None | StateIndex[tuple[Float[Array, "input_size"], Float[Array, "input_size"]]]
    )
    axis_name: Hashable | Sequence[Hashable]
    inference: bool
    input_size: int = field(static=True)
    eps: float = field(static=True)
    channelwise_affine: bool = field(static=True)
    momentum: float = field(static=True)
    mode: Literal["ema", "batch"] = field(static=True)

    def __init__(
        self,
        input_size: int,
        axis_name: Hashable | Sequence[Hashable],
        eps: float = 1e-5,
        channelwise_affine: bool = True,
        momentum: float = 0.99,
        inference: bool = False,
        dtype=None,
        mode: Literal["ema", "batch", "legacy"] = "legacy",
    ):
        """**Arguments:**

        - `input_size`: The number of channels in the input array.
        - `axis_name`: The name of the batch axis to compute statistics over, as passed
            to `axis_name` in `jax.vmap` or `jax.pmap`. Can also be a sequence (e.g. a
            tuple or a list) of names, to compute statistics over multiple named axes.
        - `eps`: Value added to the denominator for numerical stability.
        - `channelwise_affine`: Whether the module has learnable channel-wise affine
            parameters.
        - `momentum`: The rate at which to update the running statistics. Should be a
            value between 0 and 1 exclusive.
        - `inference`: If `False` then the batch means and variances will be calculated
            and used to update the running statistics. If `True` then the running
            statistics are directly used for normalisation. This may be toggled with
            [`equinox.nn.inference_mode`][] or overridden during
            [`equinox.nn.BatchNorm.__call__`][].
        - `dtype`: The dtype to use for the running statistics and the weight and bias
            if `channelwise_affine` is `True`. Defaults to either
            `jax.numpy.float32` or `jax.numpy.float64` depending on whether JAX is in
            64-bit mode.
        - `mode`: The variant of batch norm to use, either 'ema' or 'batch'.
        """
        if mode == "legacy":
            mode = "ema"
            warnings.warn(
                "When `eqx.nn.BatchNorm(..., mode=...)` is unspecified it defaults to "
                "'ema', for backward compatibility. This typically has a performance "
                "impact, and for new code the user is encouraged to use 'batch' "
                "instead. See `https://github.com/patrick-kidger/equinox/issues/659`."
            )
        if mode not in {"ema", "batch"}:
            raise ValueError("Invalid mode, must be 'ema' or 'batch'.")
        self.mode = mode
        dtype = default_floating_dtype() if dtype is None else dtype
        if channelwise_affine:
            self.weight = jnp.ones((input_size,), dtype=dtype)
            self.bias = jnp.zeros((input_size,), dtype=dtype)
        else:
            self.weight = None
            self.bias = None
        if mode == "ema":
            self.ema_first_time_index = StateIndex(jnp.array(True))
            init_buffers = (
                jnp.empty((input_size,), dtype=dtype),
                jnp.empty((input_size,), dtype=dtype),
            )
            self.ema_state_index = StateIndex(init_buffers)
            self.batch_counter = None
            self.batch_state_index = None
        else:
            self.batch_counter = StateIndex(jnp.array(0))
            init_hidden = (
                jnp.zeros((input_size,), dtype=dtype),
                jnp.ones((input_size,), dtype=dtype),
            )
            self.batch_state_index = StateIndex(init_hidden)
            self.ema_first_time_index = None
            self.ema_state_index = None
        self.inference = inference
        self.axis_name = axis_name
        self.input_size = input_size
        self.eps = eps
        self.channelwise_affine = channelwise_affine
        self.momentum = momentum

    @named_scope("eqx.nn.BatchNorm")
    def __call__(
        self,
        x: Array,
        state: State,
        *,
        key: PRNGKeyArray | None = None,
        inference: bool | None = None,
    ) -> tuple[Array, State]:
        """**Arguments:**

        - `x`: A JAX array of shape `(input_size, dim_1, ..., dim_N)`.
        - `state`: An [`equinox.nn.State`][] object (which is used to store the
            running statistics).
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)
        - `inference`: As per [`equinox.nn.BatchNorm.__init__`][]. If
            `True` or `False` then it will take priority over `self.inference`. If
            `None` then the value from `self.inference` will be used.

        **Returns:**

        A 2-tuple of:

        - A JAX array of shape `(input_size, dim_1, ..., dim_N)`.
        - An updated state object (storing the updated running statistics).

        **Raises:**

        A `NameError` if no `vmap`s are placed around this operation, or if this vmap
        does not have a matching `axis_name`.
        """
        del key

        if inference is None:
            inference = self.inference

        def _stats(y):
            mean = jnp.mean(y)
            mean = lax.pmean(mean, self.axis_name)
            var = jnp.mean((y - mean) * jnp.conj(y - mean))
            var = lax.pmean(var, self.axis_name)
            var = jnp.maximum(0.0, var)
            return mean, var

        def _norm(y, m, v, w, b):
            out = (y - m) / jnp.sqrt(v + self.eps)
            if self.channelwise_affine:
                out = out * w + b
            return out

        if self.mode == "ema":
            assert self.ema_first_time_index is not None
            assert self.ema_state_index is not None
            if inference:
                mean, var = state.get(self.ema_state_index)
            else:
                first_time = state.get(self.ema_first_time_index)
                state = state.set(self.ema_first_time_index, jnp.array(False))
                batch_mean, batch_var = jax.vmap(_stats)(x)
                running_mean, running_var = state.get(self.ema_state_index)
                momentum = self.momentum
                mean = (1 - momentum) * batch_mean + momentum * running_mean
                var = (1 - momentum) * batch_var + momentum * running_var
                # since jnp.array(0) == False
                mean = lax.select(first_time, batch_mean, mean)
                var = lax.select(first_time, batch_var, var)
                state = state.set(self.ema_state_index, (mean, var))
        else:
            assert self.batch_state_index is not None
            assert self.batch_counter is not None
            counter = state.get(self.batch_counter)
            hidden_mean, hidden_var = state.get(self.batch_state_index)
            if inference:
                # Zero-debias approach: mean = hidden_mean / (1 - momentum^counter)
                # For simplicity we do the minimal version here (no warmup).
                scale = jnp.where(counter == 0, 1, 1 - self.momentum**counter)
                mean = hidden_mean / scale
                var = hidden_var / scale
            else:
                mean, var = jax.vmap(_stats)(x)
                new_counter = counter + 1
                new_hidden_mean = hidden_mean * self.momentum + mean * (
                    1 - self.momentum
                )
                new_hidden_var = hidden_var * self.momentum + var * (1 - self.momentum)
                state = state.set(self.batch_counter, new_counter)
                state = state.set(
                    self.batch_state_index, (new_hidden_mean, new_hidden_var)
                )

        out = jax.vmap(_norm)(x, mean, var, self.weight, self.bias)
        return out, state
